# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from ._types import (
    JobBundlePurpose,  # noqa: F401
)

__all__ = ["DeadlineConfigDialog", "DeadlineLoginDialog"]

from .deadline_config_dialog import DeadlineConfigDialog
from .deadline_login_dialog import DeadlineLoginDialog



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""Module that defines the hashing algorithms supported by this library."""

import io

from enum import Enum

from ..exceptions import UnsupportedHashingAlgorithmError


class HashAlgorithm(str, Enum):
    """
    Enumerant of all hashing algorithms supported by this library.

    Algorithms:
      XXH128 - The xxhash 128-bit hashing algorithm.

    """

    XXH128 = "xxh128"


def hash_file(file_path: str, hash_alg: HashAlgorithm) -> str:
    """Hashes the given file using the given hashing algorithm."""
    if hash_alg == HashAlgorithm.XXH128:
        from xxhash import xxh3_128

        hasher = xxh3_128()
    else:
        raise UnsupportedHashingAlgorithmError(
            f"Unsupported hashing algorithm provided: {hash_alg}"
        )

    with open(file_path, "rb") as file:
        while True:
            chunk = file.read(io.DEFAULT_BUFFER_SIZE)
            if not chunk:
                break
            hasher.update(chunk)
        return hasher.hexdigest()


def hash_data(data: bytes, hash_alg: HashAlgorithm) -> str:
    """Hashes the given data bytes using the given hashing algorithm."""
    if hash_alg == HashAlgorithm.XXH128:
        from xxhash import xxh3_128

        hasher = xxh3_128()
    else:
        raise UnsupportedHashingAlgorithmError(
            f"Unsupported hashing algorithm provided: {hash_alg}"
        )

    hasher.update(data)
    return hasher.hexdigest()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
The AWS Deadline Cloud CLI interface.
"""

import logging
import sys
from logging import getLogger
from typing import Optional

import click

from .. import version
from ..api._session import session_context
from ..config import get_setting, get_setting_default
from ._common import _PROMPT_WHEN_COMPLETE

logger = getLogger(__name__)

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
_DEADLINE_LOG_LEVELS = [
    "ERROR",
    "WARNING",
    "INFO",
    "DEBUG",
]  # Log Levels AWS Deadline Cloud Allows


def _get_default_log_level() -> str:
    """
    Get the default log level from the config file.
    """
    # Set the default log level based on the setting, must do here so we can pass into the click option
    _SETTING_LOG_LEVEL = get_setting("settings.log_level").upper()
    _DEFAULT_LOG_LEVEL = get_setting_default("settings.log_level")
    _CLI_DEFAULT_LOG_LEVEL = _DEFAULT_LOG_LEVEL
    if _SETTING_LOG_LEVEL not in _DEADLINE_LOG_LEVELS:
        logger.warning(
            f"Log Level '{_SETTING_LOG_LEVEL}' not in {_DEADLINE_LOG_LEVELS}. Defaulting to {_DEFAULT_LOG_LEVEL}"
        )
    else:
        _CLI_DEFAULT_LOG_LEVEL = _SETTING_LOG_LEVEL
    return _CLI_DEFAULT_LOG_LEVEL


class ContextTrackingCommand(click.Command):
    """
    Adds the current CLI command name to User Agent headers in boto requests
    """

    def invoke(self, ctx: click.Context):
        # This is a global variable used to modify User Agent header in the default boto config
        session_context["cli-command-name"] = ctx.command_path.replace(" ", ".")
        return super().invoke(ctx)


class ContextTrackingGroup(click.Group):
    """
    Adds the current CLI command name to User Agent headers in boto requests
    """

    # Special value documented in Click to make this group class the default
    # See https://click.palletsprojects.com/en/stable/api/#click.Group.group_class
    group_class = type

    # Special value documented in Click to make this command class the default
    # See https://click.palletsprojects.com/en/stable/api/#click.Group.command_class
    command_class = ContextTrackingCommand


@click.group(cls=ContextTrackingGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version, prog_name="deadline")
@click.option(
    "--log-level",
    type=click.Choice(_DEADLINE_LOG_LEVELS, case_sensitive=False),
    default=None,
    help="Set the logging level.",
)
@click.option(
    "--redirect-output",
    help="Redirects stdout and stderr messages to append to the specified file. "
    "Useful for the 'deadlinew' command which does not produce terminal output by default on Windows.",
)
@click.option(
    "--redirect-mode",
    type=click.Choice(["append", "replace"], case_sensitive=False),
    default="append",
    help="When using the --redirect-output option, controls whether to append to or replace the output file.",
)
@click.pass_context
def main(ctx: click.Context, log_level: Optional[str], redirect_output: str, redirect_mode: str):
    """
    The AWS Deadline Cloud CLI provides functionality to interact with the AWS Deadline Cloud
    service.
    """
    if redirect_output:
        # Set both stdout and stderr to write to the specified file, writing in line buffering mode
        if redirect_mode == "append":
            open_mode = "a"
        else:
            open_mode = "w"
        sys.stdout = sys.stderr = open(redirect_output, open_mode, encoding="utf-8", buffering=1)
    if log_level is None:
        log_level = _get_default_log_level()

    logging.basicConfig(level=log_level)
    if log_level == "DEBUG":
        logger.debug("Debug logging is on")

    ctx.ensure_object(dict)
    # By default don't prompt when the operation is complete
    ctx.obj[_PROMPT_WHEN_COMPLETE] = False



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from .base_manifest import BaseAssetManifest, BaseManifestPath
from .hash_algorithms import HashAlgorithm, hash_data, hash_file
from .manifest_model import BaseManifestModel, ManifestModelRegistry
from .versions import ManifestVersion

__all__ = [
    "ManifestVersion",
    "ManifestModelRegistry",
    "BaseAssetManifest",
    "BaseManifestModel",
    "BaseManifestPath",
    "HashAlgorithm",
    "hash_data",
    "hash_file",
]

ManifestModelRegistry.register()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from __future__ import annotations
from collections import Counter
from logging import Logger, LoggerAdapter

import time
from dataclasses import asdict, dataclass, field
from enum import Enum
from threading import Lock
from typing import Callable, Dict, Optional, Union

from ._path_summarization import human_readable_file_size

CALLBACK_INTERVAL = 1  # in seconds
MAX_FILES_IN_CHUNK = 50
LOG_INTERVAL = 300  # in seconds
LOG_PERCENTAGE_THRESHOLD = 10  # in percentage


@dataclass
class SummaryStatistics:
    """
    A summary statistics metadata to be returned to the client when processing of files
    (hashing or uploading) has completed.
    The `skipped_files` refers to:
    - if this statistics is for hashing operation: the number of files whose hashing is
    skipped by the hash cache.
    - if this statistics is for uploading operation: the number of files that have already
    been uploaded to S3 bucket and thus skipped uploading.
    """

    total_time: float = 0.0  # time (in fractional seconds) taken to perform hashing or uploading
    total_files: int = 0
    total_bytes: int = 0
    processed_files: int = 0
    processed_bytes: int = 0
    skipped_files: int = 0
    skipped_bytes: int = 0
    transfer_rate: float = 0.0  # bytes/second

    def aggregate(self, other: SummaryStatistics) -> SummaryStatistics:
        """
        Aggregates other object of SummaryStatistics to this.
        """
        if not isinstance(other, self.__class__):
            raise TypeError("Only instances of the same type can be aggregated.")
        self.total_time += other.total_time
        self.total_files += other.total_files
        self.total_bytes += other.total_bytes
        self.processed_files += other.processed_files
        self.processed_bytes += other.processed_bytes
        self.skipped_files += other.skipped_files
        self.skipped_bytes += other.skipped_bytes
        self.transfer_rate = self.processed_bytes / self.total_time if self.total_time else 0.0

        return self

    def __str__(self):
        return (
            f"Processed {self.processed_files} file{'' if self.processed_files == 1 else 's'}"
            + f" totaling {human_readable_file_size(self.processed_bytes)}.\n"
            + f"Skipped re-processing {self.skipped_files} files totaling"
            + f" {human_readable_file_size(self.skipped_bytes)}.\n"
            + f"Total processing time of {round(self.total_time, ndigits=5)} seconds"
            + f" at {human_readable_file_size(int(self.transfer_rate))}/s.\n"
        )


@dataclass
class DownloadSummaryStatistics(SummaryStatistics):
    """
    A summary statistics metadata to be returned to the client when the downloading files has
    completed. In addition to the general statistics, includes a dict mapping download locations
    to the number of downloaded files in each of those locations.
    """

    file_counts_by_root_directory: Dict[str, int] = field(default_factory=dict)

    def aggregate(self, other: SummaryStatistics) -> SummaryStatistics:
        """
        Aggregates other object of DownloadSummaryStatistics to this.
        """
        super().aggregate(other)
        if not hasattr(other, "file_counts_by_root_directory"):
            raise TypeError(
                f"{other.__class__.__name__} does not have a file_counts_by_root_directory field."
            )
        else:
            self.file_counts_by_root_directory = dict(
                Counter(self.file_counts_by_root_directory)
                + Counter(other.file_counts_by_root_directory)
            )

        return self

    def convert_to_summary_statistics(self) -> SummaryStatistics:
        """
        Converts this DownloadSummaryStatistics to a SummaryStatistics.
        """
        download_summary_statistics_dict = asdict(self)
        del download_summary_statistics_dict["file_counts_by_root_directory"]
        return SummaryStatistics(**download_summary_statistics_dict)


class ProgressStatus(Enum):
    """
    Represents the current stage of asset/file processing
    """

    NONE = ("NONE", "")
    """The asset manager is not assigned any work."""

    PREPARING_IN_PROGRESS = ("PREPARING_IN_PROGRESS", "Processed")
    """The asset manager is hashing files."""

    UPLOAD_IN_PROGRESS = ("UPLOAD_IN_PROGRESS", "Uploaded")
    """The asset manager is uploading files."""

    DOWNLOAD_IN_PROGRESS = ("DOWNLOAD_IN_PROGRESS", "Downloaded")
    """Downloading files"""

    def __init__(self, title, verb_in_message):
        self.title = title
        self.verb_in_message = verb_in_message


@dataclass
class ProgressReportMetadata:
    """
    A metadata (with defined key-value pairs) about the progress to be reported
    back to client during file upload/downloads. Within this metadata will be
    a status message and progress(%) of the hashing, uploads or downloads of files.
    """

    status: ProgressStatus
    progress: float  # percentage with one decimal place
    transferRate: float  # bytes/second
    progressMessage: str  # pylint: disable=invalid-name


@dataclass
class ProgressTracker:
    """
    A class that records the progress of file processing, and reports the
    progress data back to the client using callbacks passed from the client.
    The process is one of the following - hashing, uploading, or downloading.
    """

    def __init__(
        self,
        status: ProgressStatus,
        total_files: int,
        total_bytes: int,
        on_progress_callback: Optional[Callable[[ProgressReportMetadata], bool]] = None,
        callback_interval: int = CALLBACK_INTERVAL,
        max_files_in_chunk: int = MAX_FILES_IN_CHUNK,
        logger: Optional[Union[Logger, LoggerAdapter]] = None,
        log_interval: int = LOG_INTERVAL,
        log_percentage_threshold: int = LOG_PERCENTAGE_THRESHOLD,
    ) -> None:
        def do_nothing(*args, **kwargs) -> bool:
            return True

        if not on_progress_callback:
            on_progress_callback = do_nothing

        self.on_progress_callback = on_progress_callback
        self.continue_reporting = True

        self.reporting_interval = callback_interval
        self.reporting_files_per_chunk = 1
        self.max_files_in_chunk = max_files_in_chunk
        self.completed_files_in_chunk = 0
        self.last_report_time: Optional[float] = None
        self.last_report_processed_bytes: int = 0

        self.logger = logger
        self.log_interval = log_interval
        self.log_percentage_threshold = log_percentage_threshold
        self.last_logged_time: Optional[float] = None
        self.last_logged_completed_bytes: int = 0

        self.status = status
        self.total_files = total_files
        self.total_bytes = total_bytes
        if self.total_files >= self.max_files_in_chunk:
            self.reporting_files_per_chunk = self.max_files_in_chunk
        self.processed_files = 0
        self.processed_bytes = 0
        self.skipped_files = 0
        self.skipped_bytes = 0
        self.total_time = 0.0  # total time (in fractional seconds) taken for the process

        self._lock = Lock()

        def track_progress(bytes_amount: int, current_file_done: Optional[bool] = False) -> bool:
            """
            When uploading or downloading files using boto3, pass this to the `Callback` argument
            so that the progress can be updated with the amount of bytes processed.
            """
            with self._lock:
                self._initialize_timestamps_if_none()
                self.processed_bytes += bytes_amount
                if current_file_done:
                    self.processed_files += 1
                    self.completed_files_in_chunk += 1
                # Logs progress message to the logger (if exists)
                self._log_progress_message()
                # Invokes the callback with current progress data
                return self._report_progress()

        self.track_progress_callback = track_progress

    def set_total_files(self, total_files, total_bytes) -> None:
        """
        Stores the number and size of files to be processed.
        """
        self.total_files = total_files
        self.total_bytes = total_bytes
        if self.total_files >= self.max_files_in_chunk:
            self.reporting_files_per_chunk = self.max_files_in_chunk

    def _initialize_timestamps_if_none(self) -> None:
        """
        This is to initialize the `last_report_time` and `last_logged_time` to the
        current time in alignment with the start of process (i.e., hashing, uploading,
        or downloading.) These are used to calculate the current hashing or transfer
        rate for the first progress report, which is the first callback invocation or
        the first logging.
        """
        current_time = time.perf_counter()
        if self.last_report_time is None:
            self.last_report_time = current_time
        if self.last_logged_time is None:
            self.last_logged_time = current_time

    def increase_processed(self, num_files: int = 1, file_bytes: int = 0) -> None:
        """
        Adds the number and size of processed files.
        """
        with self._lock:
            self._initialize_timestamps_if_none()
            self.processed_files += num_files
            self.completed_files_in_chunk += num_files
            self.processed_bytes += file_bytes

    def increase_skipped(self, num_files: int = 1, file_bytes: int = 0) -> None:
        """
        Adds the number and size of skipped files.
        """
        with self._lock:
            self.skipped_files += num_files
            self.completed_files_in_chunk += num_files
            self.skipped_bytes += file_bytes

    def _report_progress(self) -> bool:
        """
        Invokes the callback with current progress metadata in one of the following cases:
        1. when a specific time interval has passed since the last call, (or since the
           process started,) or
        2. when a specific number of files (a chunk) has been processed, or
        3. when the progress is 100%, (including when all files were skipped during the process.)

        Sets the flag `continue_reporting` True if the operation should continue as normal,
        or False to cancel, and returns the flag.
        """
        if not self.continue_reporting:
            return False

        current_time = time.perf_counter()
        if (
            self.last_report_time is None
            or current_time - self.last_report_time >= self.reporting_interval
            or self.completed_files_in_chunk >= self.reporting_files_per_chunk
            or self.processed_files + self.skipped_files == self.total_files
        ):
            self.continue_reporting = self.on_progress_callback(
                self._get_progress_report_metadata()
            )
            self.last_report_processed_bytes = self.processed_bytes
            self.last_report_time = current_time
            self.completed_files_in_chunk = 0
        return self.continue_reporting

    def report_progress(self) -> bool:
        with self._lock:
            return self._report_progress()

    def _get_progress_report_metadata(self) -> ProgressReportMetadata:
        completed_bytes = self.processed_bytes + self.skipped_bytes
        percentage = round(
            completed_bytes / self.total_bytes * 100 if self.total_bytes > 0 else 0, 1
        )
        seconds_since_last_report = round(
            time.perf_counter() - self.last_report_time if self.last_report_time else 0, 2
        )
        transfer_rate = (
            (self.processed_bytes - self.last_report_processed_bytes) / seconds_since_last_report
            if seconds_since_last_report > 0
            else 0
        )
        transfer_rate_name = "Transfer rate"
        if self.status == ProgressStatus.PREPARING_IN_PROGRESS:
            transfer_rate_name = "Hashing speed"

        progress_message = (
            f"{self.status.verb_in_message}"
            f" {human_readable_file_size(completed_bytes)} / {human_readable_file_size(self.total_bytes)}"
            f" of {self.total_files} file{'' if self.total_files == 1 else 's'}"
            f" ({transfer_rate_name}: {human_readable_file_size(int(transfer_rate))}/s)"
        )

        return ProgressReportMetadata(
            status=self.status,
            progress=percentage,
            transferRate=transfer_rate,
            progressMessage=progress_message,
        )

    def get_summary_statistics(self) -> SummaryStatistics:
        """
        Returns the summary statistics of hashing or upload operation.
        """
        transfer_rate = self.processed_bytes / self.total_time if self.total_time else 0.0

        return SummaryStatistics(
            total_time=self.total_time,
            total_files=self.total_files,
            total_bytes=self.total_bytes,
            processed_files=self.processed_files,
            processed_bytes=self.processed_bytes,
            skipped_files=self.skipped_files,
            skipped_bytes=self.skipped_bytes,
            transfer_rate=transfer_rate,
        )

    def get_download_summary_statistics(
        self,
        downloaded_files_paths_by_root: dict[str, list[str]],
    ) -> DownloadSummaryStatistics:
        """
        Returns the summary statistics of download operation.
        """
        summary_statistics_dict = asdict(self.get_summary_statistics())
        summary_statistics_dict["file_counts_by_root_directory"] = {
            root: len(paths) for root, paths in downloaded_files_paths_by_root.items()
        }
        return DownloadSummaryStatistics(**summary_statistics_dict)

    def _log_progress_message(self) -> None:
        """
        Logs progress message to the logger (if exists) on specific conditions:
        1. when the `log_interval` time has passed since the last call, (or since the
           process started,) or,
        2. when the tracked progress percentage difference from the last log exceeds
           the `log_percentage_threshold`, or
        3. when the process is fully completed (progress percentage reaches 100%).
        """
        if self.logger is None:
            return

        current_time = time.perf_counter()
        current_completed_bytes = self.processed_bytes + self.skipped_bytes
        progress_difference = (
            (current_completed_bytes - self.last_logged_completed_bytes) / self.total_bytes * 100
        )

        if (
            self.last_logged_time is None
            or current_time - self.last_logged_time >= self.log_interval
            or progress_difference >= self.log_percentage_threshold
            or self.processed_files + self.skipped_files == self.total_files
        ):
            self.logger.info(self._get_progress_report_metadata().progressMessage)
            self.last_logged_completed_bytes = current_completed_bytes
            self.last_logged_time = current_time



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline job` commands.
"""

from __future__ import annotations
import json
import logging
from configparser import ConfigParser
from pathlib import Path
import os
import sys
from typing import Callable, Optional, Union
import datetime
from typing import Any
import textwrap

import click
from botocore.exceptions import ClientError

from ...api._session import _modified_logging_level
from ....job_attachments.download import OutputDownloader
from ....job_attachments.models import (
    FileConflictResolution,
    JobAttachmentS3Settings,
    PathFormat,
)
from ....job_attachments._utils import (
    WINDOWS_MAX_PATH_LENGTH,
    _is_windows_long_path_registry_enabled,
)
from ....job_attachments.progress_tracker import (
    DownloadSummaryStatistics,
    ProgressReportMetadata,
)
from ....job_attachments._path_summarization import (
    human_readable_file_size,
    summarize_path_list,
)

from ... import api
from ...config import config_file
from ...exceptions import DeadlineOperationError, DeadlineOperationTimedOut
from .._common import _apply_cli_options_to_config, _cli_object_repr, _handle_error
from .._main import main
from ._sigint_handler import SigIntHandler
from ...api._session import get_default_client_config

logger = logging.getLogger("deadline.client.cli")

JSON_MSG_TYPE_TITLE = "title"
JSON_MSG_TYPE_PRESUMMARY = "presummary"
JSON_MSG_TYPE_PATH = "path"
JSON_MSG_TYPE_PATHCONFIRM = "pathconfirm"
JSON_MSG_TYPE_PROGRESS = "progress"
JSON_MSG_TYPE_SUMMARY = "summary"
JSON_MSG_TYPE_ERROR = "error"
JSON_MSG_TYPE_WARNING = "warning"


def _format_timestamp(timestamp: datetime.datetime, use_local_time: bool = False) -> str:
    """
    Format a timestamp in ISO 8601 format with timezone information.

    Args:
        timestamp: The datetime object to format
        use_local_time: If True, convert to local time; if False, keep as UTC

    Returns:
        Formatted timestamp string in ISO 8601 format with timezone
    """
    if use_local_time and timestamp.tzinfo is not None:
        # Convert to local time
        local_timestamp = timestamp.astimezone()
        return local_timestamp.isoformat()
    else:
        # Keep as UTC - ensure it has timezone info
        if timestamp.tzinfo is None:
            # Assume naive datetime is UTC
            utc_timestamp = timestamp.replace(tzinfo=datetime.timezone.utc)
        else:
            utc_timestamp = timestamp
        return utc_timestamp.isoformat()


# Set up the signal handler for handling Ctrl + C interruptions.
sigint_handler = SigIntHandler()


@main.group(name="job")
@_handle_error
def cli_job():
    """
    Commands to work with jobs.
    """


@cli_job.command(name="list")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--page-size", default=5, help="The number of jobs to load at a time.")
@click.option("--item-offset", default=0, help="The index of the job to start listing from.")
@_handle_error
def job_list(page_size, item_offset, **args):
    """
    Lists the Jobs in a queue.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    try:
        response = deadline.search_jobs(
            farmId=farm_id,
            queueIds=[queue_id],
            itemOffset=item_offset,
            pageSize=page_size,
            sortExpressions=[{"fieldSort": {"name": "CREATED_AT", "sortOrder": "DESCENDING"}}],
        )
    except ClientError as exc:
        raise DeadlineOperationError(f"Failed to get Jobs from Deadline:\n{exc}") from exc

    total_results = response["totalResults"]

    # Select which fields to print and in which order
    name_field = "displayName"
    if len(response["jobs"]) and "name" in response["jobs"][0]:
        name_field = "name"
    structured_job_list = [
        {
            field: job.get(field, "")
            for field in [
                name_field,
                "jobId",
                "taskRunStatus",
                "startedAt",
                "endedAt",
                "createdBy",
                "createdAt",
            ]
        }
        for job in response["jobs"]
    ]

    click.echo(
        f"Displaying {len(structured_job_list)} of {total_results} Jobs starting at {item_offset}"
    )
    click.echo()
    click.echo(_cli_object_repr(structured_job_list))


@cli_job.command(name="get")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to get.")
@_handle_error
def job_get(**args):
    """
    Get the details of a job.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    response = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    response.pop("ResponseMetadata", None)

    click.echo(_cli_object_repr(response))


@cli_job.command(name="cancel")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to cancel.")
@click.option(
    "--mark-as",
    type=click.Choice(["SUSPENDED", "CANCELED", "FAILED", "SUCCEEDED"], case_sensitive=False),
    default="CANCELED",
    help="The status to apply to all active tasks in the job.",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@_handle_error
def job_cancel(mark_as: str, yes: bool, **args):
    """
    Cancel job from running.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    mark_as = mark_as.upper()

    deadline = api.get_boto3_client("deadline", config=config)

    # Print a summary of the job to cancel
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    # Remove the zero-count status counts
    job["taskRunStatusCounts"] = {
        name: count for name, count in job["taskRunStatusCounts"].items() if count != 0
    }
    # Filter the fields to a summary
    filtered_job = {
        field: job.get(field, "")
        for field in [
            "name",
            "jobId",
            "taskRunStatus",
            "taskRunStatusCounts",
            "startedAt",
            "endedAt",
            "createdBy",
            "createdAt",
        ]
    }
    click.echo(_cli_object_repr(filtered_job))

    # Ask for confirmation about canceling this job.
    if not (
        yes or config_file.str2bool(config_file.get_setting("settings.auto_accept", config=config))
    ):
        if mark_as == "CANCELED":
            cancel_message = "Are you sure you want to cancel this job?"
        else:
            cancel_message = (
                f"Are you sure you want to cancel this job and mark its taskRunStatus as {mark_as}?"
            )
        # We explicitly require a yes/no response, as this is an operation that will interrupt the work in progress
        # on their job.
        if not click.confirm(
            cancel_message,
            default=None,
        ):
            click.echo("Job not canceled.")
            sys.exit(1)

    if mark_as == "CANCELED":
        click.echo("Canceling job...")
    else:
        click.echo(f"Canceling job and marking as {mark_as}...")
    deadline.update_job(farmId=farm_id, queueId=queue_id, jobId=job_id, targetTaskRunStatus=mark_as)


@cli_job.command(name="requeue-tasks")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to requeue tasks for.")
@click.option(
    "--run-status",
    type=click.Choice(
        ["SUSPENDED", "CANCELED", "FAILED", "SUCCEEDED", "NOT_COMPATIBLE"], case_sensitive=False
    ),
    multiple=True,
    help="Requeue tasks of this status. Repeat the option to provide multiple statuses.",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@_handle_error
def job_requeue_tasks(run_status: Optional[list[str]], **args):
    """
    Requeue tasks of a job. By default, requeues all FAILED, CANCELED, and SUSPENDED tasks.

    Use the --run-status option to requeue tasks of different status.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    if not run_status:
        run_status_set = {"SUSPENDED", "CANCELED", "FAILED"}
    else:
        run_status_set = {status.upper() for status in run_status}

    session = api.get_boto3_session(config=config)
    deadline_client = api.get_boto3_client("deadline", config=config)

    # Create a separate API client for the update_task calls, using the "adaptive" retry strategy.
    # This strategy is better suited to repeatedly calling the API for a longer duration, because it
    # retains backoff data across calls instead of starting fresh each time.
    requeues_client_config = get_default_client_config(
        retries=dict(mode="adaptive", total_max_attempts=5)
    )
    deadline_client_for_requeues = session.client("deadline", config=requeues_client_config)

    # Print a summary of the job's task run statuses
    job = deadline_client.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)

    click.echo(f"Job: {job['name']} ({job['jobId']})")
    # Remove the zero-count status counts for a shorter summary
    task_run_status_counts = {
        name.upper(): count for name, count in job["taskRunStatusCounts"].items() if count != 0
    }
    click.echo(_cli_object_repr({"taskRunStatusCounts": task_run_status_counts}))
    click.echo(f"Requeuing all tasks with run status among: {', '.join(sorted(run_status_set))}")

    total_task_count_to_requeue = sum(
        count for name, count in task_run_status_counts.items() if name in run_status_set
    )
    summary_task_count_by_status = ", ".join(
        f"{count} {name} tasks"
        for name, count in task_run_status_counts.items()
        if name in run_status_set and count != 0
    )
    summary_tasks_message = f"{total_task_count_to_requeue} total tasks"

    if total_task_count_to_requeue == 0:
        click.echo("No tasks to requeue.")
        return

    # Ask for confirmation about requeuing the selected tasks
    if config_file.str2bool(config_file.get_setting("settings.auto_accept", config=config)):
        click.echo(
            f"Estimated {summary_tasks_message} ({summary_task_count_by_status}) to requeue."
        )
    else:
        click.echo(
            f"This action will requeue an estimated {summary_tasks_message} ({summary_task_count_by_status})"
        )
        if not click.confirm(
            "Are you sure you want to requeue these tasks?",
            default=None,
        ):
            click.echo("No tasks were requeued.")
            sys.exit(1)
        click.echo("Requeuing tasks...")

    total_count_requeued = 0
    # Use a paginator to get all the steps
    paginator = deadline_client.get_paginator("list_steps")
    steps = []
    for page in paginator.paginate(farmId=farm_id, queueId=queue_id, jobId=job_id):
        steps.extend(page["steps"])

    # For each step, requeue all the tasks matching the run statuses
    for step in steps:
        click.echo(f"\nStep: {step['name']} ({step['stepId']})")
        step_task_count_to_requeue = sum(
            count for name, count in step["taskRunStatusCounts"].items() if name in run_status_set
        )
        summary_task_count_by_status = ", ".join(
            f"{count} {name} tasks"
            for name, count in step["taskRunStatusCounts"].items()
            if name in run_status_set and count != 0
        )
        summary_tasks_message = f"{step_task_count_to_requeue} total tasks"
        if step_task_count_to_requeue == 0:
            click.echo("  Step has no tasks to requeue.")
            continue
        else:
            click.echo(
                f"  Requeuing an estimated {summary_tasks_message} ({summary_task_count_by_status})..."
            )

        paginator = deadline_client.get_paginator("list_tasks")
        for page in paginator.paginate(
            farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step["stepId"]
        ):
            for task in page["tasks"]:
                if task["runStatus"].upper() in run_status_set:
                    parameters = task.get("parameters", {})
                    if parameters:
                        task_summary = (
                            ",".join(
                                f"{param}={list(parameters[param].values())[0]}"
                                for param in parameters
                            )
                            + f" ({task['taskId']})"
                        )
                    else:
                        task_summary = task["taskId"]
                    click.echo(f"    {task['runStatus']} {task_summary}")
                    # Requeue means to set the target run status to PENDING. If a task has no dependencies,
                    # it will switch to READY immediately.
                    deadline_client_for_requeues.update_task(
                        farmId=farm_id,
                        queueId=queue_id,
                        jobId=job_id,
                        stepId=step["stepId"],
                        taskId=task["taskId"],
                        targetRunStatus="PENDING",
                    )
                    total_count_requeued += 1

    click.echo(f"\nRequeued a total of {total_count_requeued} tasks.")


def _download_job_output(
    config: Optional[ConfigParser],
    farm_id: str,
    queue_id: str,
    job_id: str,
    step_id: Optional[str],
    task_id: Optional[str],
    is_json_format: bool = False,
):
    """
    Starts the download of job output and handles the progress reporting callback.
    """
    deadline = api.get_boto3_client("deadline", config=config)

    auto_accept = config_file.str2bool(
        config_file.get_setting("settings.auto_accept", config=config)
    )
    conflict_resolution = config_file.get_setting("settings.conflict_resolution", config=config)

    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    step = {}
    task = {}
    if step_id:
        step = deadline.get_step(farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step_id)
    if task_id:
        task = deadline.get_task(
            farmId=farm_id,
            queueId=queue_id,
            jobId=job_id,
            stepId=step_id,
            taskId=task_id,
        )

    click.echo(
        _get_start_message(job["name"], step.get("name"), task.get("parameters"), is_json_format)
    )

    queue = deadline.get_queue(farmId=farm_id, queueId=queue_id)

    queue_role_session = api.get_queue_user_boto3_session(
        deadline=deadline,
        config=config,
        farm_id=farm_id,
        queue_id=queue_id,
        queue_display_name=queue["displayName"],
    )

    # Get a dictionary mapping rootPath to rootPathFormat (OS) from job's manifests
    root_path_format_mapping: dict[str, str] = {}
    job_attachments = job.get("attachments", None)
    if job_attachments:
        job_attachments_manifests = job_attachments["manifests"]
        for manifest in job_attachments_manifests:
            root_path_format_mapping[manifest["rootPath"]] = manifest["rootPathFormat"]

    job_output_downloader = OutputDownloader(
        s3_settings=JobAttachmentS3Settings(**queue["jobAttachmentSettings"]),
        farm_id=farm_id,
        queue_id=queue_id,
        job_id=job_id,
        step_id=step_id,
        task_id=task_id,
        session=queue_role_session,
    )

    def _check_and_warn_long_output_paths(
        output_paths_by_root: dict[str, list[str]],
    ) -> None:
        if sys.platform == "win32" and not _is_windows_long_path_registry_enabled():
            for root, paths in output_paths_by_root.items():
                for output_path in paths:
                    if len(root + output_path) >= WINDOWS_MAX_PATH_LENGTH:
                        click.secho(
                            _get_long_path_found_message(is_json_format),
                            fg="yellow",
                        )

    output_paths_by_root = job_output_downloader.get_output_paths_by_root()
    # If no output paths were found, log a message and exit.
    if output_paths_by_root == {}:
        click.echo(_get_no_output_message(is_json_format))
        return

    _check_and_warn_long_output_paths(output_paths_by_root)

    # Check if the asset roots came from different OS. If so, prompt users to
    # select alternative root paths to download to, (regardless of the auto-accept.)
    asset_roots = list(output_paths_by_root.keys())
    for asset_root in asset_roots:
        root_path_format = root_path_format_mapping.get(asset_root, "")
        if root_path_format == "":
            # There must be a corresponding root path format for each root path, by design.
            raise DeadlineOperationError(f"No root path format found for {asset_root}.")
        if PathFormat.get_host_path_format_string() != root_path_format:
            click.echo(_get_mismatch_os_root_warning(asset_root, root_path_format, is_json_format))

            if not is_json_format:
                new_root = click.prompt(
                    "> Please enter a new root path",
                    type=click.Path(exists=False),
                )
            else:
                json_string = click.prompt("", prompt_suffix="", type=str)
                new_root = _get_value_from_json_line(
                    json_string, JSON_MSG_TYPE_PATHCONFIRM, expected_size=1
                )[0]
                _assert_valid_path(new_root)

            job_output_downloader.set_root_path(asset_root, os.path.expanduser(new_root))

    output_paths_by_root = job_output_downloader.get_output_paths_by_root()

    _check_and_warn_long_output_paths(output_paths_by_root)

    # Prompt users to confirm local root paths where they will download outputs to,
    # and allow users to select different location to download files to if they want.
    # (If auto-accept is enabled, automatically download to the default root paths.)
    if not auto_accept:
        if not is_json_format:
            user_choice = ""
            while user_choice != ("y" or "n"):
                click.echo(
                    _get_summary_of_files_to_download_message(output_paths_by_root, is_json_format)
                )
                asset_roots = list(output_paths_by_root.keys())
                click.echo(_get_roots_list_message(asset_roots, is_json_format))
                user_choice = click.prompt(
                    "> Please enter the index of root directory to edit, y to proceed without changes, or n to cancel the download",
                    type=click.Choice(
                        [
                            *[str(num) for num in list(range(0, len(asset_roots)))],
                            "y",
                            "n",
                        ]
                    ),
                    default="y",
                )
                if user_choice == "n":
                    click.echo("Output download canceled.")
                    return
                elif user_choice != "y":
                    # User selected an index to modify the root directory.
                    index_to_change = int(user_choice)
                    new_root = click.prompt(
                        "> Please enter the new root directory path, or press Enter to keep it unchanged",
                        type=click.Path(exists=False),
                        default=asset_roots[index_to_change],
                    )
                    job_output_downloader.set_root_path(
                        asset_roots[index_to_change], str(Path(new_root))
                    )
                    output_paths_by_root = job_output_downloader.get_output_paths_by_root()
                    _check_and_warn_long_output_paths(output_paths_by_root)
        else:
            click.echo(
                _get_summary_of_files_to_download_message(output_paths_by_root, is_json_format)
            )
            asset_roots = list(output_paths_by_root.keys())
            click.echo(_get_roots_list_message(asset_roots, is_json_format))
            json_string = click.prompt("", prompt_suffix="", type=str)
            confirmed_asset_roots = _get_value_from_json_line(
                json_string, JSON_MSG_TYPE_PATHCONFIRM, expected_size=len(asset_roots)
            )
            for index, confirmed_root in enumerate(confirmed_asset_roots):
                _assert_valid_path(confirmed_root)
                job_output_downloader.set_root_path(asset_roots[index], str(Path(confirmed_root)))
            output_paths_by_root = job_output_downloader.get_output_paths_by_root()
            _check_and_warn_long_output_paths(output_paths_by_root)

    if not is_json_format:
        # Create and print a summary of all the paths to download
        all_output_paths: set[str] = set()
        for asset_root, output_paths in output_paths_by_root.items():
            all_output_paths.update(
                os.path.normpath(os.path.join(asset_root, path)) for path in output_paths
            )
        click.echo("\nSummary of file paths to download:")
        click.echo(textwrap.indent(summarize_path_list(all_output_paths), "  "))

    # If the conflict resolution option was not specified, auto-accept is false, and
    # if there are any conflicting files in local, prompt users to select a resolution method.
    # (skip, overwrite, or make a copy.)
    if conflict_resolution != FileConflictResolution.NOT_SELECTED.name:
        file_conflict_resolution = FileConflictResolution[conflict_resolution]
    elif auto_accept:
        file_conflict_resolution = FileConflictResolution.CREATE_COPY
    else:
        file_conflict_resolution = FileConflictResolution.CREATE_COPY
        conflicting_filenames = _get_conflicting_filenames(output_paths_by_root)
        if conflicting_filenames:
            click.echo(_get_conflict_resolution_selection_message(conflicting_filenames))
            user_choice = click.prompt(
                "> Please enter your choice (1, 2, 3, or n to cancel the download)",
                type=click.Choice(["1", "2", "3", "n"]),
                default="3",
            )
            if user_choice == "n":
                click.echo("Output download canceled.")
                return
            else:
                resolution_choice_int = int(user_choice)
                file_conflict_resolution = FileConflictResolution(resolution_choice_int)

    # TODO: remove logging level setting when the max number connections for boto3 client
    # in Job Attachments library can be increased (currently using default number, 10, which
    # makes it keep logging urllib3 warning messages when downloading large files)
    with _modified_logging_level(logging.getLogger("urllib3"), logging.ERROR):

        @api.record_success_fail_telemetry_event(metric_name="download_job_output")
        def _download_job_output(
            file_conflict_resolution: Optional[
                FileConflictResolution
            ] = FileConflictResolution.CREATE_COPY,
            on_downloading_files: Optional[Callable[[ProgressReportMetadata], bool]] = None,
        ) -> DownloadSummaryStatistics:
            return job_output_downloader.download_job_output(
                file_conflict_resolution=file_conflict_resolution,
                on_downloading_files=on_downloading_files,
            )

        if not is_json_format:
            # Note: click doesn't export the return type of progressbar(), so we suppress mypy warnings for
            # not annotating the type of download_progress.
            with click.progressbar(length=100, label="Downloading Outputs") as download_progress:  # type: ignore[var-annotated]

                def _update_download_progress(
                    download_metadata: ProgressReportMetadata,
                ) -> bool:
                    new_progress = int(download_metadata.progress) - download_progress.pos
                    if new_progress > 0:
                        download_progress.update(new_progress)
                    return sigint_handler.continue_operation

                download_summary: DownloadSummaryStatistics = _download_job_output(  # type: ignore
                    file_conflict_resolution=file_conflict_resolution,
                    on_downloading_files=_update_download_progress,
                )
        else:

            def _update_download_progress(
                download_metadata: ProgressReportMetadata,
            ) -> bool:
                click.echo(
                    _get_json_line(JSON_MSG_TYPE_PROGRESS, str(int(download_metadata.progress)))
                )
                # TODO: enable download cancellation for JSON format
                return True

            download_summary = _download_job_output(  # type: ignore
                file_conflict_resolution=file_conflict_resolution,
                on_downloading_files=_update_download_progress,
            )

    click.echo(_get_download_summary_message(download_summary, is_json_format))
    click.echo()


def _get_start_message(
    job_name: str,
    step_name: Optional[str],
    task_parameters: Optional[dict],
    is_json_format: bool,
) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_TITLE, job_name)
    else:
        if step_name is None:
            return f"Downloading output from Job {job_name!r}"
        elif task_parameters is None:
            return f"Downloading output from Job {job_name!r} Step {step_name!r}"
        else:
            task_parameters_summary = "{}"
            if task_parameters:
                task_parameters_summary = (
                    "{"
                    + ",".join(
                        f"{key}={list(value.values())[0]}" for key, value in task_parameters.items()
                    )
                    + "}"
                )
            return f"Downloading output from Job {job_name!r} Step {step_name!r} Task {task_parameters_summary}"


def _get_no_output_message(is_json_format: bool) -> str:
    msg = (
        "There are no output files available for download at this moment. Please verify that"
        " the Job/Step/Task you are trying to download output from has completed successfully."
    )
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_SUMMARY, msg)
    else:
        return msg


def _get_mismatch_os_root_warning(root: str, root_path_format: str, is_json_format: bool) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PATH, [root])
    else:
        path_format_capitalized_first_letter = root_path_format[0].upper() + root_path_format[1:]
        return (
            "This root path format does not match the operating system you're using. "
            "Where would you like to save the files?\n"
            f"The location was {root}, on {path_format_capitalized_first_letter}."
        )


def _get_summary_of_files_to_download_message(
    output_paths_by_root: dict[str, list[str]], is_json_format: bool
) -> str:
    # Print some information about what we will download
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PRESUMMARY, output_paths_by_root)
    else:
        paths_message_joined = "    " + "\n    ".join(
            f"{os.path.commonpath([os.path.join(directory, p) for p in output_paths])} ({len(output_paths)} file{'s' if len(output_paths) > 1 else ''})"
            for directory, output_paths in output_paths_by_root.items()
        )
        return f"\nSummary of files to download:\n{paths_message_joined}\n"


def _get_roots_list_message(asset_roots: list[str], is_json_format: bool) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PATH, asset_roots)
    else:
        asset_roots_str = "\n".join([f"[{index}] {root}" for index, root in enumerate(asset_roots)])
        return (
            f"You are about to download files which may come from multiple root directories. Here are a list of the current root directories:\n"
            f"{asset_roots_str}"
        )


def _get_conflict_resolution_selection_message(conflicting_filenames: list[str]) -> str:
    conflicting_filenames_str = "\n        ".join(conflicting_filenames)
    return (
        f"The following files already exist in your local directory:\n        {conflicting_filenames_str}\n"
        f"You have three options to choose from:\n"
        f"[1] Skip: Do not download these files\n"
        f"[2] Overwrite: Download these files and overwrite existing files\n"
        f"[3] Create a copy: Download the file with a new name, appending '(1)' to the end"
    )


def _get_download_summary_message(
    download_summary: DownloadSummaryStatistics, is_json_format: bool
) -> str:
    if is_json_format:
        return _get_json_line(
            JSON_MSG_TYPE_SUMMARY,
            f"Downloaded {download_summary.processed_files} files",
            extra_properties={"fileCount": download_summary.processed_files},
        )
    else:
        paths_joined = "\n        ".join(
            f"{directory} ({count} file{'s' if count > 1 else ''})"
            for directory, count in download_summary.file_counts_by_root_directory.items()
        )
        return (
            "Download Summary:\n"
            f"    Downloaded {download_summary.processed_files} files totaling"
            f" {human_readable_file_size(download_summary.processed_bytes)}.\n"
            f"    Total download time of {round(download_summary.total_time, ndigits=5)} seconds"
            f" at {human_readable_file_size(int(download_summary.transfer_rate))}/s.\n"
            f"    Download locations (total file counts):\n        {paths_joined}"
        )


def _get_long_path_found_message(is_json_format: bool) -> str:
    message = """
WARNING: Found downloaded file paths that exceed Windows path length limit. This may cause unexpected issues.
For details and a fix using the registry, see: https://learn.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
"""
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_WARNING, message)

    else:
        return message


def _get_conflicting_filenames(filenames_by_root: dict[str, list[str]]) -> list[str]:
    conflicting_filenames: list[str] = []

    for root, filenames in filenames_by_root.items():
        for filename in filenames:
            abs_path = Path(root).joinpath(filename).resolve()
            if abs_path.is_file():
                conflicting_filenames.append(str(abs_path))

    return conflicting_filenames


def _get_json_line(
    messageType: str,
    value: Union[str, list[str], dict[str, Any]],
    extra_properties: Optional[dict[str, Any]] = None,
) -> str:
    json_obj = {"messageType": messageType, "value": value}
    if extra_properties:
        json_obj.update(extra_properties)
    return json.dumps(json_obj, ensure_ascii=True)


def _get_value_from_json_line(
    json_line: str, message_type: str, expected_size: Optional[int] = None
) -> Union[str, list[str]]:
    try:
        parsed_json = json.loads(json_line)
        if parsed_json["messageType"] != message_type:
            raise ValueError(
                f"Expected message type '{message_type}' but received '{parsed_json['messageType']}'"
            )
        if expected_size and len(parsed_json["value"]) != expected_size:
            raise ValueError(
                f"Expected {expected_size} item{'' if expected_size == 1 else 's'} in value "
                f"but received {len(parsed_json['value'])}"
            )
        return parsed_json["value"]
    except Exception as e:
        raise ValueError(f"Invalid JSON line '{json_line}': {e}")


def _assert_valid_path(path: str) -> None:
    """
    Validates that the path has the format of the OS currently running.
    """
    path_obj = Path(path)
    if not path_obj.is_absolute():
        raise ValueError(f"Path {path} is not an absolute path.")


@cli_job.command(name="download-output")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to use.")
@click.option("--step-id", help="The step to use.")
@click.option("--task-id", help="The task to use.")
@click.option(
    "--conflict-resolution",
    type=click.Choice(
        [
            FileConflictResolution.SKIP.name,
            FileConflictResolution.OVERWRITE.name,
            FileConflictResolution.CREATE_COPY.name,
        ],
        case_sensitive=False,
    ),
    help="How to handle downloads if a file already exists:\n"
    "CREATE_COPY (default): Download the file with a new name, appending '(1)' to the end\n"
    "SKIP: Do not download the file\n"
    "OVERWRITE: Download and replace the existing file",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Automatically accept any confirmation prompts",
)
@click.option(
    "--output",
    type=click.Choice(
        ["verbose", "json"],
        case_sensitive=False,
    ),
    help="Specifies the output format of the messages printed to stdout.\n"
    "VERBOSE: Displays messages in a human-readable text format.\n"
    "JSON: Displays messages in JSON line format, so that the info can be easily "
    "parsed/consumed by custom scripts.",
)
@_handle_error
def job_download_output(step_id, task_id, output, **args):
    """
    Download a job's output.
    """
    if task_id and not step_id:
        raise click.UsageError("Missing option '--step-id' required with '--task-id'")
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)
    is_json_format = True if output == "json" else False

    try:
        _download_job_output(config, farm_id, queue_id, job_id, step_id, task_id, is_json_format)
    except Exception as e:
        if is_json_format:
            error_one_liner = str(e).replace("\n", ". ")
            click.echo(_get_json_line(JSON_MSG_TYPE_ERROR, error_one_liner))
            sys.exit(1)
        else:
            if logging.DEBUG >= logger.getEffectiveLevel():
                logger.exception("Exception details:")
            raise DeadlineOperationError(f"Failed to download output:\n{e}") from e


@cli_job.command(name="wait")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to wait for.")
@click.option("--max-poll-interval", default=120, help="Maximum polling interval in seconds.")
@click.option("--timeout", default=0, help="Timeout in seconds (0 for no timeout).")
@click.option(
    "--output",
    type=click.Choice(["verbose", "json"], case_sensitive=False),
    default="verbose",
    help="Output format (verbose or json).",
)
@_handle_error
def job_wait_for_completion(max_poll_interval, timeout, output, **args):
    """
    Wait for a job to complete and return failed step-task IDs.

    This command blocks until the job's taskRunStatus reaches a terminal state (SUCCEEDED, FAILED, CANCELED, SUSPENDED, or NOT_COMPATIBLE),
    then returns a list of any failed step-task combinations.

    The command uses exponential backoff for polling, starting at 0.5 seconds and doubling
    the interval after each check until it reaches the maximum polling interval.

    Exit codes:
    0 - Job succeeded
    1 - Timeout waiting for job completion
    2 - Job failed (any tasks failed)
    3 - Job was canceled
    4 - Job was suspended
    5 - Job is not compatible
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    is_json_output = output.lower() == "json"

    # Get job name for output
    deadline = api.get_boto3_client("deadline", config=config)
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    job_name = job["name"]

    # Define a status callback for verbose output
    def status_callback(status, elapsed_time=0, total_timeout=0):
        if not is_json_output:
            timeout_info = ""
            if total_timeout > 0:
                remaining = max(0, total_timeout - elapsed_time)
                timeout_info = f" [{elapsed_time:.1f}s elapsed, {remaining:.1f}s remaining]"
            else:
                timeout_info = f" [{elapsed_time:.1f}s elapsed]"

            # Clear the current line and update in place
            click.echo(
                f"\rCurrent status: {status}.{timeout_info}",
                nl=False,
            )

    if not is_json_output:
        click.echo(f"Waiting for job {job_id} to complete...")

    try:
        result = api.wait_for_job_completion(
            farm_id=farm_id,
            queue_id=queue_id,
            job_id=job_id,
            max_poll_interval=max_poll_interval,
            timeout=timeout,
            config=config,
            status_callback=status_callback,
        )

        if is_json_output:
            # Return everything as JSON
            response = {
                "jobId": job_id,
                "jobName": job_name,
                "status": result.status,
                "elapsedTime": result.elapsed_time,
                "failedTasks": [
                    {
                        "stepId": task.step_id,
                        "taskId": task.task_id,
                        "stepName": task.step_name,
                        "sessionId": task.session_id,
                    }
                    for task in result.failed_tasks
                ],
            }
            click.echo(json.dumps(response, indent=2))
        else:
            # Use verbose output with YAML formatting
            click.echo(f"Job ID: {job_id}")
            click.echo(f"Job Name: {job_name}")
            click.echo(f"Job completed with status: {result.status}")
            click.echo(f"Elapsed time: {result.elapsed_time:.1f} seconds")

            if result.failed_tasks:
                click.echo(f"Found {len(result.failed_tasks)} failed tasks:")
                failed_tasks_dict = [
                    {
                        "stepId": task.step_id,
                        "taskId": task.task_id,
                        "stepName": task.step_name,
                        "sessionId": task.session_id,
                    }
                    for task in result.failed_tasks
                ]
                click.echo(_cli_object_repr(failed_tasks_dict))
            else:
                click.echo("No failed tasks found.")

        # Determine exit code based on job status
        exit_code = 0
        if result.status == "SUCCEEDED" and not result.failed_tasks:
            exit_code = 0
        elif result.status == "FAILED" or result.failed_tasks:
            exit_code = 2
        elif result.status == "CANCELED":
            exit_code = 3
        elif result.status == "SUSPENDED":
            exit_code = 4
        elif result.status == "ARCHIVED":
            exit_code = 4
        elif result.status == "NOT_COMPATIBLE":
            exit_code = 5
        else:
            # Unknown status, treat as failure
            exit_code = 2

        sys.exit(exit_code)

    except DeadlineOperationTimedOut as e:
        if is_json_output:
            error_response = {
                "error": str(e),
                "timeout": True,
                "jobId": job_id,
                "jobName": job_name,
            }
            click.echo(json.dumps(error_response, indent=2))
        else:
            click.echo(f"Job ID: {job_id}")
            click.echo(f"Job Name: {job_name}")
            click.echo(f"Error waiting for job completion: {e}")
        sys.exit(1)
    except DeadlineOperationError as e:
        if is_json_output:
            error_response = {
                "error": str(e),
                "jobId": job_id,
                "jobName": job_name,
            }
            click.echo(json.dumps(error_response, indent=2))
        else:
            click.echo(f"Job ID: {job_id}")
            click.echo(f"Job Name: {job_name}")
            click.echo(f"Error waiting for job completion: {e}")
        sys.exit(2)


@cli_job.command(name="logs")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to get logs for.")
@click.option(
    "--session-id",
    help="The session ID to get logs for. If not provided and job-id is specified, will use the latest session based on endedAt time.",
)
@click.option("--limit", default=100, help="Maximum number of log lines to return.")
@click.option(
    "--start-time",
    help="Start time for logs in ISO format (e.g., 2023-01-01T12:00:00Z).",
)
@click.option("--end-time", help="End time for logs in ISO format (e.g., 2023-01-01T13:00:00Z).")
@click.option("--next-token", help="Token for pagination of results.")
@click.option(
    "--output",
    type=click.Choice(["verbose", "json"], case_sensitive=False),
    default="verbose",
    help="Output format (verbose or json).",
)
@click.option(
    "--timezone",
    type=click.Choice(["utc", "local"], case_sensitive=False),
    default="utc",
    help="Timezone for timestamps (utc or local). Default is utc.",
)
@_handle_error
def job_logs(session_id, limit, start_time, end_time, next_token, output, timezone, **args):
    """
    Get CloudWatch logs for a specific session.

    This command retrieves logs from CloudWatch for the specified session ID.
    By default, it returns the most recent 100 log lines, but this can be
    adjusted using the --limit parameter.

    If session-id is not provided but job-id is, the command will automatically
    select a session using the following priority:
    1. If there are ongoing sessions (no endedAt time), always prefer them
    2. Among ongoing sessions, select the one that started most recently
    3. If no ongoing sessions exist, select the completed session that ended most recently

    Use --next-token with the value from a previous response to get the next page of results.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    is_json_output = output.lower() == "json"

    # Get job name if job_id is available
    deadline = api.get_boto3_client("deadline", config=config)
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    job_name = job["name"]

    use_local_time = timezone.lower() == "local"

    # If session_id is not provided but job_id is, try to find the session
    if not session_id and job_id:
        try:
            # Use paginator to get all sessions
            paginator = deadline.get_paginator("list_sessions")
            sessions = []

            for page in paginator.paginate(farmId=farm_id, queueId=queue_id, jobId=job_id):
                sessions.extend(page.get("sessions", []))

            if not sessions:
                raise DeadlineOperationError(f"No sessions found for job {job_id}")
            elif len(sessions) == 1:
                session_id = sessions[0]["sessionId"]
                if not is_json_output:
                    click.echo(f"Using the only available session: {session_id}")
            else:
                # Multiple sessions found, select the latest one
                # Prioritize ongoing sessions (no endedAt) over completed ones
                # Among ongoing sessions, select the one that started most recently
                # Among completed sessions, select the one that ended most recently
                ongoing_sessions = [s for s in sessions if "endedAt" not in s]
                completed_sessions = [s for s in sessions if "endedAt" in s]

                if ongoing_sessions:
                    # Always prefer ongoing sessions, select the most recently started one
                    latest_session = max(
                        ongoing_sessions,
                        key=lambda s: s.get(
                            "startedAt", datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
                        ),
                    )
                else:
                    # No ongoing sessions, select the most recently completed one
                    latest_session = max(
                        completed_sessions,
                        key=lambda s: s.get(
                            "endedAt", datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
                        ),
                    )

                session_id = latest_session["sessionId"]
                if not is_json_output:
                    click.echo(f"Using the latest session: {session_id}")
        except ClientError as exc:
            raise DeadlineOperationError(
                f"Failed to list sessions for job {job_id}:\n{exc}"
            ) from exc

    # Ensure we have a session ID at this point
    if not session_id:
        raise DeadlineOperationError(
            "Session ID is required. Provide it with --session-id or specify a --job-id with at least one session."
        )

    if not is_json_output:
        click.echo(
            f"Retrieving logs for session {session_id} from log group /aws/deadline/{farm_id}/{queue_id}..."
        )

    try:
        result = api.get_session_logs(
            farm_id=farm_id,
            queue_id=queue_id,
            session_id=session_id,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
            next_token=next_token,
            config=config,
        )

        if is_json_output:
            # Return everything as JSON
            response = {
                "jobId": job_id,
                "jobName": job_name,
                "events": [
                    {
                        "timestamp": _format_timestamp(event.timestamp, use_local_time),
                        "message": event.message,
                        "ingestionTime": (
                            _format_timestamp(event.ingestion_time, use_local_time)
                            if event.ingestion_time
                            else None
                        ),
                        "eventId": event.event_id,
                    }
                    for event in result.events
                ],
                "count": result.count,
                "nextToken": result.next_token,
                "logGroup": result.log_group,
                "logStream": result.log_stream,
            }
            click.echo(json.dumps(response, indent=2))
        else:
            # Display the logs in verbose format
            if not result.events:
                click.echo("No logs found for the specified session.")
                return

            # Display job information if available
            click.echo(f"Job ID: {job_id}")
            click.echo(f"Job Name: {job_name}")

            for event in result.events:
                timestamp = _format_timestamp(event.timestamp, use_local_time)
                click.echo(f"[{timestamp}] {event.message}")

            click.echo(f"\nRetrieved {result.count} log events.")

            # If there are more logs available, inform the user
            if result.next_token:
                click.echo(
                    f'More logs are available. Use --next-token "{result.next_token}" to retrieve the next page.'
                )

    except Exception as e:
        if is_json_output:
            error_response = {"error": str(e)}
            click.echo(json.dumps(error_response, indent=2))
            sys.exit(1)
        else:
            raise DeadlineOperationError(f"Error retrieving logs: {e}")


@cli_job.command(name="trace-schedule")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to trace.")
@click.option("-v", "--verbose", is_flag=True, help="Output verbose trace details.")
@click.option(
    "--trace-format",
    type=click.Choice(
        ["chrome"],
        case_sensitive=False,
    ),
    help="The tracing format to write.",
)
@click.option("--trace-file", help="The tracing file to write.")
@_handle_error
def job_trace_schedule(verbose, trace_format, trace_file, **args):
    """
    EXPERIMENTAL - Generate statistics from a job.

    To visualize the trace output file when providing the options
    "--trace-format chrome --trace-file <output>.json", use
    the https://ui.perfetto.dev Tracing UI and choose "Open trace file".
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    if trace_file and not trace_format:
        raise DeadlineOperationError("Error: Must provide --trace-format with --trace-file.")

    deadline = api.get_boto3_client("deadline", config=config)
    trace_end_utc = datetime.datetime.now(datetime.timezone.utc)

    click.echo("Getting the job...")
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    job.pop("ResponseMetadata", None)

    if "startedAt" not in job:
        raise DeadlineOperationError("No trace available - Job hasn't started yet, exiting")
    started_at = job["startedAt"]

    click.echo("Getting all the sessions for the job...")
    response = deadline.list_sessions(farmId=farm_id, queueId=queue_id, jobId=job_id)
    while "nextToken" in response:
        old_list = response["sessions"]
        response = deadline.list_sessions(
            farmId=farm_id,
            queueId=queue_id,
            jobId=job_id,
            nextToken=response["nextToken"],
        )
        response["sessions"] = old_list + response["sessions"]
    response.pop("ResponseMetadata", None)

    sessions = sorted(response["sessions"], key=lambda session: session["startedAt"])

    click.echo("Getting all the session actions for the job...")
    for session in sessions:
        response = deadline.list_session_actions(
            farmId=farm_id,
            queueId=queue_id,
            jobId=job_id,
            sessionId=session["sessionId"],
        )
        while "nextToken" in response:
            old_list = response["sessionActions"]
            response = deadline.list_session_actions(
                farmId=farm_id,
                queueId=queue_id,
                jobId=job_id,
                sessionId=session["sessionId"],
                nextToken=response["nextToken"],
            )
            response["sessionActions"] = old_list + response["sessionActions"]
        response.pop("ResponseMetadata", None)

        session["actions"] = response["sessionActions"]

    # Cache steps and tasks by their id, to only get each once
    steps: dict[str, Any] = {}
    tasks: dict[str, Any] = {}

    with click.progressbar(  # type: ignore[var-annotated]
        length=len(sessions), label="Getting all the steps and tasks for the job..."
    ) as progressbar:
        for index, session in enumerate(sessions):
            session["index"] = index
            for action in session["actions"]:
                step_id = action["definition"].get("taskRun", {}).get("stepId")
                task_id = action["definition"].get("taskRun", {}).get("taskId")
                if step_id and task_id:
                    if "step" not in session:
                        if step_id in steps:
                            step = steps[step_id]
                        else:
                            step = deadline.get_step(
                                farmId=farm_id,
                                queueId=queue_id,
                                jobId=job_id,
                                stepId=step_id,
                            )
                            step.pop("ResponseMetadata", None)
                            steps[step_id] = step
                        session["step"] = step
                    elif session["step"]["stepId"] != step_id:
                        # The session itself doesn't have a step id, but for now the scheduler always creates new
                        # sessions for new steps.
                        raise DeadlineOperationError(
                            f"Session {session['sessionId']} ran more than one step! When this code was"
                            " written that wasn't possible."
                        )

                    if task_id in tasks:
                        task = tasks[task_id]
                    else:
                        task = deadline.get_task(
                            farmId=farm_id,
                            queueId=queue_id,
                            jobId=job_id,
                            stepId=step_id,
                            taskId=task_id,
                        )
                        task.pop("ResponseMetadata", None)
                        tasks[task_id] = task
                    action["task"] = task
            progressbar.update(1)

    # Collect the worker IDs that ran the sessions, and give them indexes to act as PIDs in the tracing file
    worker_ids = {session["workerId"] for session in sessions}
    workers = {worker_id: index for index, worker_id in enumerate(worker_ids)}

    click.echo("Processing the trace data...")
    trace_events = []

    def time_int(timestamp: datetime.datetime):
        return int((timestamp - started_at) / datetime.timedelta(microseconds=1))

    def duration_of(resource):
        try:
            return time_int(resource.get("endedAt", trace_end_utc)) - time_int(
                resource["startedAt"]
            )
        except KeyError:
            return 0

    accumulators = {
        "sessionCount": 0,
        "sessionActionCount": 0,
        "taskRunCount": 0,
        "envActionCount": 0,
        "syncJobAttachmentsCount": 0,
        "sessionDuration": 0,
        "sessionActionDuration": 0,
        "taskRunDuration": 0,
        "envActionDuration": 0,
        "syncJobAttachmentsDuration": 0,
    }

    for session in sessions:
        accumulators["sessionCount"] += 1
        accumulators["sessionDuration"] += duration_of(session)

        pid = workers[session["workerId"]]
        session_event_name = f"{session['step']['name']} - {session['index']}"
        if "endedAt" not in session:
            session_event_name = f"{session_event_name} - In Progress"
        trace_events.append(
            {
                "name": session_event_name,
                "cat": "SESSION",
                "ph": "B",  # Begin Event
                "ts": time_int(session["startedAt"]),
                "pid": pid,
                "tid": 0,
                "args": {
                    "sessionId": session["sessionId"],
                    "workerId": session["workerId"],
                    "fleetId": session["fleetId"],
                    "lifecycleStatus": session["lifecycleStatus"],
                },
            }
        )

        for action in session["actions"]:
            accumulators["sessionActionCount"] += 1
            accumulators["sessionActionDuration"] += duration_of(action)

            name = action["sessionActionId"]
            action_type = list(action["definition"].keys())[0]
            if action_type == "taskRun":
                accumulators["taskRunCount"] += 1
                accumulators["taskRunDuration"] += duration_of(action)

                task = action["task"]
                parameters = task.get("parameters", {})
                name = ",".join(
                    f"{param}={list(parameters[param].values())[0]}" for param in parameters
                )
                if not name:
                    name = "<No Task Params>"
            elif action_type in ("envEnter", "envExit"):
                accumulators["envActionCount"] += 1
                accumulators["envActionDuration"] += duration_of(action)

                name = action["definition"][action_type]["environmentId"].split(":")[-1]
            elif action_type == "syncInputJobAttachments":
                accumulators["syncJobAttachmentsCount"] += 1
                accumulators["syncJobAttachmentsDuration"] += duration_of(action)

                if "stepId" in action["definition"][action_type]:
                    name = "Sync Job Attchmnt (Dependencies)"
                else:
                    name = "Sync Job Attchmnt (Submitted)"
            if "endedAt" not in action:
                name = f"{name} - In Progress"
            if "startedAt" in action:
                trace_events.append(
                    {
                        "name": name,
                        "cat": action_type,
                        "ph": "X",  # Complete Event
                        "ts": time_int(action["startedAt"]),
                        "dur": duration_of(action),
                        "pid": pid,
                        "tid": 0,
                        "args": {
                            "sessionActionId": action["sessionActionId"],
                            "status": action["status"],
                            "stepName": session["step"]["name"],
                        },
                    }
                )

        trace_events.append(
            {
                "name": session_event_name,
                "cat": "SESSION",
                "ph": "E",  # End Event
                "ts": time_int(session.get("endedAt", trace_end_utc)),
                "pid": pid,
                "tid": 0,
            }
        )

    if verbose:
        click.echo(" ==== TRACE DATA ====")
        click.echo(_cli_object_repr(job))
        click.echo("")
        click.echo(_cli_object_repr(sessions))

    click.echo("")
    click.echo(" ==== SUMMARY ====")
    click.echo("")
    click.echo(f"Session Count: {accumulators['sessionCount']}")
    session_total_duration = accumulators["sessionDuration"]
    click.echo(f"Session Total Duration: {datetime.timedelta(microseconds=session_total_duration)}")
    click.echo(f"Session Action Count: {accumulators['sessionActionCount']}")
    click.echo(
        f"Session Action Total Duration: {datetime.timedelta(microseconds=accumulators['sessionActionDuration'])}"
    )
    click.echo(f"Task Run Count: {accumulators['taskRunCount']}")
    task_run_total_duration = accumulators["taskRunDuration"]
    click.echo(
        f"Task Run Total Duration: {datetime.timedelta(microseconds=task_run_total_duration)} ({100 * task_run_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(
        f"Non-Task Run Count: {accumulators['sessionActionCount'] - accumulators['taskRunCount']}"
    )
    non_task_run_total_duration = (
        accumulators["sessionActionDuration"] - accumulators["taskRunDuration"]
    )
    click.echo(
        f"Non-Task Run Total Duration: {datetime.timedelta(microseconds=non_task_run_total_duration)} ({100 * non_task_run_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(f"Sync Job Attachments Count: {accumulators['syncJobAttachmentsCount']}")
    sync_job_attachments_total_duration = accumulators["syncJobAttachmentsDuration"]
    click.echo(
        f"Sync Job Attachments Total Duration: {datetime.timedelta(microseconds=sync_job_attachments_total_duration)} ({100 * sync_job_attachments_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(f"Env Action Count: {accumulators['envActionCount']}")
    env_action_total_duration = accumulators["envActionDuration"]
    click.echo(
        f"Env Action Total Duration: {datetime.timedelta(microseconds=env_action_total_duration)} ({100 * env_action_total_duration / session_total_duration:.1f}%)"
    )
    click.echo("")
    within_session_overhead_duration = (
        accumulators["sessionDuration"] - accumulators["sessionActionDuration"]
    )
    click.echo(
        f"Within-session Overhead Duration: {datetime.timedelta(microseconds=within_session_overhead_duration)} ({100 * within_session_overhead_duration / session_total_duration:.1f}%)"
    )
    click.echo(
        f"Within-session Overhead Duration Per Action: {datetime.timedelta(microseconds=(accumulators['sessionDuration'] - accumulators['sessionActionDuration']) / accumulators['sessionActionCount'])}"
    )

    tracing_data: dict[str, Any] = {
        "traceEvents": trace_events,
        # "displayTimeUnits": "s",
        "otherData": {
            "farmId": farm_id,
            "queueId": queue_id,
            "jobId": job_id,
            "jobName": job["name"],
            "startedAt": job["startedAt"].isoformat(sep="T"),
        },
    }
    if "endedAt" in job:
        tracing_data["otherData"]["endedAt"] = job["endedAt"].isoformat(sep="T")

    tracing_data["otherData"].update(accumulators)

    if trace_file:
        with open(trace_file, "w", encoding="utf8") as f:
            json.dump(tracing_data, f, indent=1)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from __future__ import annotations

import os
import shutil

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from typing import Any


class HatchCustomBuildHook(BuildHookInterface):
    """
    This class implements Hatch's [custom build hook] (https://hatch.pypa.io/1.6/plugins/build-hook/custom/)
    for a copy_version_py operation that copies the _version.py file generated by the hatch-vcs build hook into
    specified destination directories. See the `[[tool.hatch.build.hooks.custom]]` section in `pyproject.toml`.
    """

    def _validate_config(self):
        if sorted(self.config) != ["copy_version_py", "path"] or list(
            self.config["copy_version_py"]
        ) != ["destinations"]:
            raise RuntimeError(
                "Configuration of the custom build hook must be like { 'copy_version_py': {'destinations': ['path1', ...]}}."
                + f" Received:\n{self.config}"
            )

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        self._validate_config()

        for destination in self.config["copy_version_py"]["destinations"]:
            print(f"Copying _version.py to {destination}")
            shutil.copy(
                os.path.join(self.root, "_version.py"),
                os.path.join(self.root, destination),
            )

    def clean(self, versions: list[str]) -> None:
        self._validate_config()

        cleaned_count = 0
        for destination in self.config["copy_version_py"]["destinations"]:
            print(f"Cleaning _version.py from {destination}")
            clean_path = os.path.join(self.root, destination, "_version.py")
            try:
                os.remove(clean_path)
                cleaned_count += 1
            except FileNotFoundError:
                pass
        print(f"Cleaned {cleaned_count} items")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
UI widgets for the host requirements tab.
"""

import re
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

from qtpy.QtCore import Qt  # type: ignore
from qtpy.QtGui import (  # type: ignore
    QBrush,
    QDoubleValidator,
    QFont,
    QIcon,
    QIntValidator,
    QRegularExpressionValidator,
    QStandardItem,
    QValidator,
)
from qtpy.QtWidgets import (  # type: ignore
    QComboBox,
    QDoubleSpinBox,
    QFrame,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListView,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QRadioButton,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from deadline.client.exceptions import NonValidInputError

from ..dataclasses import CustomRequirements, HardwareRequirements, HostRequirements, OsRequirements

logger = getLogger(__name__)

MAX_INT_VALUE = (2**31) - 1
MIN_INT_VALUE = -(2**31) + 1
DECIMAL_VALUE = 2
LABEL_FIXED_WIDTH: int = 150
BUTTON_FIXED_WIDTH: int = 150

ATTRIBUTE = "Attribute"
AMOUNT = "Amount"
PLACEHOLDER_TEXT = "-"
INFO_ICON_PATH = str(Path(__file__).parent.parent / "resources" / "info.svg")
CUSTOM_REQUIREMENT_TOOL_TIP = (
    "<html>"
    "<p>"
    "<b>Custom worker requirements</b><br>"
    "With this feature, you can define your own custom worker "
    "capabilities. There are two kinds of worker capabilities you can add."
    "</p>"
    "<ul>"
    "<li><b>Amount</b> - you can define the quantity of something that the worker needs to have for "
    "a step to run. For example, you might define the number of licenses required.</li>"
    "<li><b>Attribute</b> - You can define a property or attribute the worker needs for a step to run. "
    "The attributes are always defined as a set of strings. For example, for a software configuration, "
    "you might define it as 'SoftwareConfig = Option1'.</li>"
    "</ul>"
    "</html>"
)

CUSTOM_CAPABILITY_NAME_REGEX = "^(\\.[a-zA-Z][a-zA-Z0-9]{0,63})+$"
ATTRIBUTE_CAPABILITY_VALUE_REGEX = "^[a-zA-Z_]([a-zA-Z0-9_\\-]{0,99})$"

ATTRIBUTE_CAPABILITY_PREFIX = "attr."
AMOUNT_CAPABILITY_PREFIX = "amount."

RESERVED_FIRST_IDENTIFIERS = ["worker", "job", "step", "task"]

# An attribute name needs to be <= 100 characters. Each Identifier must start with a letter or underscore and can
# be a maximum of 64 characters long. Periods separate each Identifier.
IDENTIFIER_REGEX = "[a-zA-Z_][a-zA-Z0-9_]{0,63}"

ATTRIBUTE_CAPABILITY_NAME_REGEX = rf"^({IDENTIFIER_REGEX})(\.{IDENTIFIER_REGEX})*$"
AMOUNT_CAPABILITY_NAME_REGEX = rf"^({IDENTIFIER_REGEX})(\.{IDENTIFIER_REGEX})*$"


class AddIcon(QIcon):
    def __init__(self):
        file_path = str(Path(__file__).parent.parent / "resources" / "add.svg")
        super().__init__(file_path)


class HostRequirementsWidget(QWidget):  # pylint: disable=too-few-public-methods
    """
    UI Elements that hold host requirements settings across all job types.

    Args:
        parent: The parent Qt Widget.
    """

    def __init__(self, requirements: Optional[HostRequirements] = None, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout(self)

        self.mode_selection_box = OverrideRequirementsWidget(self)
        layout.addWidget(self.mode_selection_box)

        self.os_requirements_box = OSRequirementsWidget(self)
        self.os_requirements_box.setEnabled(False)
        layout.addWidget(self.os_requirements_box)

        self.hardware_requirements_box = HardwareRequirementsWidget(self)
        self.hardware_requirements_box.setEnabled(False)
        layout.addWidget(self.hardware_requirements_box)

        self.custom_requirements_box = CustomRequirementsWidget(self)
        self.custom_requirements_box.setEnabled(False)
        layout.addWidget(self.custom_requirements_box)

        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        self.mode_selection_box.use_custom_button.toggled.connect(
            self._on_mode_selection_use_custom_button_toggled
        )
        if requirements is not None:
            self.set_requirements(requirements)

    def _on_mode_selection_use_custom_button_toggled(self, state):
        """
        Enable all settings when use_custom_button is selected. Otherwise disable all settings.
        """
        self.os_requirements_box.setEnabled(state)
        self.hardware_requirements_box.setEnabled(state)
        self.custom_requirements_box.setEnabled(state)

    def _is_custom_requirements_selected(self) -> bool:
        return self.mode_selection_box.use_custom_button.isChecked()

    def get_requirements(self) -> Optional[Dict[str, Any]]:
        """
        Returns a list of OpenJD parameter definition dicts with values filled from the widget.
        If requirement settings are not enabled, then return None.

        host_requirements: dict[str, Any] = {
           "amounts": [ <AmountRequirement>, ... ], # @optional
           "attributes": [ <AttributeRequirement>, ... ] # @optional
        }
        """
        if not self._is_custom_requirements_selected():
            return None

        os_requirements = self.os_requirements_box.get_requirements()
        hardware_requirements = self.hardware_requirements_box.get_requirements()
        custom_requirements = self.custom_requirements_box.get_requirements()

        requirements: Dict[str, Union[List[str], List[int]]] = {}
        if os_requirements:
            requirements.setdefault("attributes", []).extend(os_requirements)  # type: ignore
        if hardware_requirements:
            requirements.setdefault("amounts", []).extend(hardware_requirements)  # type: ignore
        if custom_requirements["amounts"]:
            requirements.setdefault("amounts", []).extend(custom_requirements["amounts"])  # type: ignore
        if custom_requirements["attributes"]:
            requirements.setdefault("attributes", []).extend(custom_requirements["attributes"])  # type: ignore

        return requirements

    def set_requirements(self, requirements: HostRequirements):
        if requirements.custom_requirements:
            self.custom_requirements_box.set_requirements(requirements.custom_requirements)
        if requirements.os_requirements:
            self.os_requirements_box.set_requirements(requirements.os_requirements)
        if requirements.hardware_requirements:
            self.hardware_requirements_box.set_requirements(requirements.hardware_requirements)
        self.mode_selection_box.use_default_button.setChecked(False)
        self.mode_selection_box.use_custom_button.setChecked(True)


class OverrideRequirementsWidget(QGroupBox):  # pylint: disable=too-few-public-methods
    """
    UI elements to hold top level selection between using all workers or selected workers that meet requirements.

    Args:
        parent: The parent Qt Widget.
    """

    def __init__(self, parent=None):
        super().__init__("", parent)
        self.layout = QVBoxLayout(self)
        self._build_ui()

    def _build_ui(self):
        # Use Fleet Default button
        self.use_default_button = QRadioButton("Run on all available worker hosts")
        self.use_default_button.setChecked(True)

        # Use customized settings button + tip
        self.use_custom_button = QRadioButton(
            "Run on worker hosts that meet the following requirements"
        )
        self.use_custom_button_tip = QHBoxLayout()
        self.custom_button_tip_text = QLabel("All fields below are optional")
        custom_button_label_font = self.custom_button_tip_text.font()
        custom_button_label_font.setPointSize(10)
        custom_button_label_font.setItalic(True)
        custom_button_label_font.setWeight(QFont.Light)
        self.custom_button_tip_text.setFont(custom_button_label_font)
        self.custom_button_tip_text.setWordWrap(True)
        self.custom_button_tip_text.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.custom_button_tip_text.setAlignment(Qt.AlignTop)

        # account for 2pt spacing left/right of radio button
        self.use_custom_button_tip.addSpacing(self.use_custom_button.iconSize().width() + 4)
        self.use_custom_button_tip.addWidget(self.custom_button_tip_text)

        # Stack the components
        self.layout.addWidget(self.use_default_button)
        self.layout.addSpacing(6)
        self.layout.addWidget(self.use_custom_button)
        self.layout.addLayout(self.use_custom_button_tip)


class OSRequirementsWidget(QGroupBox):
    """
    UI elements to hold OS settings.

    Args:
        parent: The parent Qt Widget.
    """

    OS_ROW_OPTIONS = ["linux", "macos", "windows"]
    CPU_ROW_OPTIONS = ["x86_64", "arm64"]

    def __init__(self, parent=None):
        super().__init__("", parent=parent)
        self.layout = QVBoxLayout(self)
        self._build_ui()

    def _build_ui(self):
        self.os_row = OSRequirementRowWidget("Operating system", self.OS_ROW_OPTIONS)
        self.cpu_row = OSRequirementRowWidget("CPU architecture", self.CPU_ROW_OPTIONS)

        self.layout.addWidget(self.os_row)
        self.layout.addWidget(self.cpu_row)

    def get_requirements(self) -> List[Dict[str, Any]]:
        """
        Returns a list of OpenJD parameter definition dicts with
        a "value" key filled from the widget.

        Set the following capabilities according to OpenJD spec.
        - attr.worker.os.family
        - attr.worker.cpu.arch
        """

        # TODO: currently only supports "AnyOf" from the UI
        requirements: List[dict] = []
        if self.os_row.combo_box.checkedItems():
            requirements.append(
                {
                    "name": "attr.worker.os.family",
                    "anyOf": self.os_row.combo_box.checkedItems(),
                }
            )
        if self.cpu_row.combo_box.checkedItems():
            requirements.append(
                {
                    "name": "attr.worker.cpu.arch",
                    "anyOf": self.cpu_row.combo_box.checkedItems(),
                }
            )
        return requirements

    def set_requirements(self, requirements: OsRequirements):
        self.os_row.combo_box.setChecked(requirements.operating_systems)  # type: ignore
        self.cpu_row.combo_box.setChecked(requirements.cpu_archs)  # type: ignore


class HardwareRequirementsWidget(QGroupBox):  # pylint: disable=too-few-public-methods
    """
    UI elements to hold list of hardware requirements.

    Args:
        parent: The parent Qt Widget.
    """

    def __init__(self, parent=None):
        super().__init__("Hardware requirements", parent)
        self.layout = QVBoxLayout(self)
        self._build_ui()

    def _build_ui(self):
        # Build a custom row widget for each selectable option
        self.cpu_row = HardwareRequirementsRowWidget("vCPUs", self)
        self.memory_row = HardwareRequirementsRowWidget("Memory (GiB)", self)
        self.gpu_row = HardwareRequirementsRowWidget("GPUs", self)
        self.gpu_memory_row = HardwareRequirementsRowWidget("GPU memory (GiB)", self)
        self.scratch_space_row = HardwareRequirementsRowWidget("Scratch space", self)

        # Add all rows to layout
        self.layout.addWidget(self.cpu_row)
        self.layout.addWidget(self.memory_row)
        self.layout.addWidget(self.gpu_row)
        self.layout.addWidget(self.gpu_memory_row)
        self.layout.addWidget(self.scratch_space_row)

    def get_requirements(self) -> List[Dict[str, Any]]:
        """
        Returns a list of OpenJD parameter definition dicts with
        a "value" key filled from the widget.

        Set the following capabilities according to OpenJD spec.
        - amount.worker.vcpu
        - amount.worker.memory
        - amount.worker.gpu
        - amount.worker.gpu.memory
        - amount.worker.disk.scratch
        """
        requirements: List[Dict[str, Any]] = []
        self.cpu_row.add_requirement(requirements, "amount.worker.vcpu")
        # Memory capability has UI unit in GiB but template unit in MiB, so setting scaling factor to 1024
        self.memory_row.add_requirement(requirements, "amount.worker.memory", 1024)
        self.gpu_row.add_requirement(requirements, "amount.worker.gpu")
        # GPU Memory capability has UI unit in GiB but template unit in MiB, so set scaling factor to 1024
        self.gpu_memory_row.add_requirement(requirements, "amount.worker.gpu.memory", 1024)
        # Disk Scratch capability has unit in GiB
        self.scratch_space_row.add_requirement(requirements, "amount.worker.disk.scratch")
        return requirements

    def set_requirements(self, requirements: HardwareRequirements):
        self.cpu_row.clear()
        self.memory_row.clear()
        self.gpu_row.clear()
        self.gpu_memory_row.clear()
        self.scratch_space_row.clear()

        self.cpu_row.minimum = requirements.cpu_min
        self.cpu_row.maximum = requirements.cpu_max

        self.memory_row.minimum = requirements.memory_min // 1024
        self.memory_row.maximum = requirements.memory_max // 1024

        self.gpu_row.minimum = requirements.acceleration_min
        self.gpu_row.maximum = requirements.acceleration_max

        self.gpu_memory_row.minimum = requirements.acceleration_memory_min // 1024
        self.gpu_memory_row.maximum = requirements.acceleration_memory_max // 1024

        self.scratch_space_row.minimum = requirements.scratch_space_min
        self.scratch_space_row.maximum = requirements.scratch_space_max


class CustomRequirementsWidget(QGroupBox):
    """
    UI elements to hold custom host requirements.

    Args:
        parent: The parent Qt Widget.
    """

    def __init__(self, parent=None):
        super().__init__("Custom host requirements", parent)
        self.layout = QVBoxLayout(self)
        self.attribute_index_numbers = set()
        self.amount_index_numbers = set()
        self._build_ui()

    def _build_ui(self):
        # Add a label that will display tool tip when hovered above
        self.info = QLabel(
            f"<html><img src={INFO_ICON_PATH} width='10' height='10'> More info</html>"
        )
        info_font = self.info.font()
        info_font.setPointSize(10)
        self.info.setFont(info_font)
        self.info.setToolTip(CUSTOM_REQUIREMENT_TOOL_TIP)

        self.info_row = QHBoxLayout()
        self.info_row.setAlignment(Qt.AlignLeft)
        self.info_row.addWidget(self.info)

        # Create a list widget for placing custom capability items
        # - no frame & no background
        # - disable directly selecting list items
        # - no scroll bars
        self.list_widget = QListWidget(self)
        self.list_widget.setSpacing(2)
        self.list_widget.setSelectionMode(QListView.NoSelection)
        self.list_widget.viewport().setAutoFillBackground(False)
        self.list_widget.setFrameStyle(QFrame.NoFrame)
        self.list_widget.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.list_widget.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.list_widget.setSizeAdjustPolicy(QListWidget.AdjustToContents)
        self.resize_list_to_fit()

        # Add a row with Add Amount and Add Attribute buttons
        self.add_amount_button = QPushButton("Add amount")
        self.add_amount_button.setFixedWidth(BUTTON_FIXED_WIDTH)

        self.add_attr_button = QPushButton("Add attribute")
        self.add_attr_button.setFixedWidth(BUTTON_FIXED_WIDTH)

        self.buttons_row = QHBoxLayout()
        self.buttons_row.setAlignment(Qt.AlignLeft)
        self.buttons_row.addWidget(self.add_amount_button)
        self.buttons_row.addWidget(self.add_attr_button)

        self.add_amount_button.clicked.connect(self._add_new_custom_amount)
        self.add_attr_button.clicked.connect(self._add_new_custom_attr)

        # Add everything together
        self.layout.addLayout(self.info_row)
        self.layout.addWidget(self.list_widget)
        self.layout.addLayout(self.buttons_row)

    def _add_new_custom_amount(self):  # type: ignore
        # Return type CustomAmountWidget is not yet defined
        return self._add_new_item(AMOUNT)

    def _add_new_custom_attr(self):  # type: ignore
        # Return type CustomAttributeWidget is not yet defined
        return self._add_new_item(ATTRIBUTE)

    def _add_new_item(self, type):
        list_item = QListWidgetItem(self.list_widget)

        if type == ATTRIBUTE:
            new_attribute_number = max(self.attribute_index_numbers, default=0) + 1
            item = CustomAttributeWidget(list_item, new_attribute_number, self)
            self.attribute_index_numbers.add(new_attribute_number)
        elif type == AMOUNT:
            new_amount_number = max(self.amount_index_numbers, default=0) + 1
            item = CustomAmountWidget(list_item, new_amount_number, self)
            self.amount_index_numbers.add(new_amount_number)
        else:
            raise NonValidInputError(f"Unexpected item type when adding new item: {type}")

        list_item.setSizeHint(item.sizeHint())

        self.list_widget.addItem(list_item)
        self.list_widget.setItemWidget(list_item, item)
        self.resize_list_to_fit()
        return item

    def remove_widget_item(self, custom_capability_widget):
        # remove the ListWidgetItem from list
        if custom_capability_widget.capability_type == ATTRIBUTE:
            self.attribute_index_numbers.remove(custom_capability_widget.item_number)
        elif custom_capability_widget.capability_type == AMOUNT:
            self.amount_index_numbers.remove(custom_capability_widget.item_number)
        else:
            raise NonValidInputError(
                f"Unexpected item type when removing item: {custom_capability_widget.capability_type}"
            )

        item = custom_capability_widget.list_item
        self.list_widget.takeItem(self.list_widget.indexFromItem(item).row())

        self.resize_list_to_fit()

    def resize_list_to_fit(self):
        # Resize the list widget to based on the size of the contents
        if self.list_widget.count() == 0:
            self.list_widget.setFixedHeight(0)
        else:
            current_height = 0
            for i in range(self.list_widget.count()):
                widget = self.list_widget.itemWidget(self.list_widget.item(i))
                if widget is not None:
                    current_height += widget.height() + 2 * self.list_widget.spacing()

            self.list_widget.setFixedHeight(current_height + 2 * self.list_widget.frameWidth())

    def get_requirements(self) -> Dict[str, List]:
        """
        Returns two lists of OpenJD parameter definition dicts
        for both amounts and attributes requirements.
        """
        requirements: Dict[str, Any] = {"amounts": [], "attributes": []}
        for i in range(self.list_widget.count()):
            widget = self.list_widget.itemWidget(self.list_widget.item(i))
            widget_requirement = widget.get_requirement()
            if widget_requirement:
                if isinstance(widget, CustomAmountWidget):
                    requirements["amounts"].append(widget_requirement)
                elif isinstance(widget, CustomAttributeWidget):
                    requirements["attributes"].append(widget_requirement)
                else:
                    logger.warning(
                        f"Widget requirement is not a valid expected type: {type(widget)}"
                    )

        return requirements

    def set_requirements(self, requirements: CustomRequirements):
        self.list_widget.clear()

        for requirement in requirements.amounts:
            amount_widget: CustomAmountWidget = self._add_new_custom_amount()
            amount_widget.name = requirement.name
            amount_widget.minimum = requirement.min
            amount_widget.maximum = requirement.max

        for requirement in requirements.attributes:
            attribute_widget: CustomAttributeWidget = self._add_new_custom_attr()
            attribute_widget.name = requirement.name
            attribute_widget.option = requirement.option
            attribute_widget.values = requirement.values


class CustomCapabilityWidget(QGroupBox):
    """
    UI element to hold a single custom requirement, either Attribute or Amount.
    """

    def __init__(
        self,
        capability_type: str,
        list_item: QListWidgetItem,
        item_number: int,
        parent: CustomRequirementsWidget,
    ):
        super().__init__(parent)
        self._parent: CustomRequirementsWidget = parent
        self.capability_type = capability_type
        self.list_item = list_item
        self.item_number = item_number

        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.title_label = QLabel(f"{capability_type} {item_number}")
        self.title_label.setStyleSheet("font-weight: bold")

        self.delete_button = QPushButton("Delete")
        self.delete_button.clicked.connect(self._delete)

        self.title_row = QHBoxLayout()
        self.title_row.addWidget(self.title_label)
        self.title_row.addStretch()
        self.title_row.addWidget(self.delete_button)

        self.layout.addLayout(self.title_row)

    def _delete(self):
        self._parent.remove_widget_item(self)
        self.setParent(None)
        self.deleteLater()


class CustomAmountWidget(CustomCapabilityWidget):
    """
    UI element to hold a single custom attribute.
    """

    def __init__(self, list_item: QListWidgetItem, item_number: int, parent=None):
        super().__init__(AMOUNT, list_item, item_number, parent)
        self._build_ui()

    def _build_ui(self):
        # Name / Value
        self.name_label = QLabel("Amount name")
        self.name_label.setFixedWidth(LABEL_FIXED_WIDTH)
        self.name_line_edit = QLineEdit()
        self.name_line_edit.setFixedWidth(LABEL_FIXED_WIDTH)
        self.name_line_edit.setValidator(QRegularExpressionValidator(AMOUNT_CAPABILITY_NAME_REGEX))
        assert (100 - len(AMOUNT_CAPABILITY_PREFIX)) > 0
        self.name_line_edit.setMaxLength(100 - len(AMOUNT_CAPABILITY_PREFIX))

        # Create layout with min/max spinbox
        self.min_label = QLabel("Min")
        self.min_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.max_label = QLabel("Max")
        self.max_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.min_spin_box = OptionalDoubleSpinBox(
            min=0, max=MAX_INT_VALUE, decimal=DECIMAL_VALUE, parent=self
        )
        self.min_spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.max_spin_box = OptionalDoubleSpinBox(
            min=0, max=MAX_INT_VALUE, decimal=DECIMAL_VALUE, parent=self
        )
        self.max_spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.min_max_row = QHBoxLayout()
        self.min_max_row.addWidget(self.min_label)
        self.min_max_row.addWidget(self.min_spin_box)
        self.min_max_row.addWidget(self.max_label)
        self.min_max_row.addWidget(self.max_spin_box)

        self.name_column = QVBoxLayout()
        self.name_column.setContentsMargins(2, 0, 0, 0)
        self.name_column.addWidget(self.name_label)
        self.name_column.addWidget(self.name_line_edit)

        self.value_column = QVBoxLayout()
        self.value_column.setContentsMargins(0, 0, 0, 0)
        self.value_column.addStretch()
        self.value_column.addLayout(self.min_max_row)

        self.columns = QHBoxLayout()
        self.columns.setContentsMargins(0, 0, 0, 15)

        self.columns.addLayout(self.name_column)
        self.columns.addLayout(self.value_column)

        # LineEdit  / LineEdit / Optional [X]
        self.layout.addLayout(self.columns)

    @property
    def name(self) -> str:
        return self.name_line_edit.text()

    @name.setter
    def name(self, name: str):
        self.name_line_edit.setText(name)

    @property
    def minimum(self) -> Optional[float]:
        if self.min_spin_box.has_input():
            return self.min_spin_box.value()
        return None

    @minimum.setter
    def minimum(self, minimum: int):
        self.min_spin_box.setValue(minimum)

    @property
    def maximum(self) -> Optional[float]:
        if self.max_spin_box.has_input():
            return self.max_spin_box.value()
        return None

    @maximum.setter
    def maximum(self, maximum: int):
        self.max_spin_box.setValue(maximum)

    def get_requirement(self) -> Dict[str, Any]:
        """
        Returns an OpenJD parameter definition dict with
        a "value" key filled from the widget.

        An amount capability is prefixed with "amount.worker.".
        """
        requirement: Dict[str, Any] = {}
        if self.name_line_edit.text():
            if self.name[-1] == ".":
                raise NonValidInputError("Your requirement name cannot end with a period.")

            requirement = {"name": AMOUNT_CAPABILITY_PREFIX + self.name}
            minimum = self.minimum
            maximum = self.maximum

            if minimum and maximum:
                requirement["min"] = minimum
                requirement["max"] = maximum

                if minimum > maximum:
                    raise NonValidInputError(
                        "Please make sure that the custom amounts in the custom host requirement options have valid min/max ranges!"
                    )
            elif minimum:
                requirement["min"] = minimum
            elif maximum:
                requirement["max"] = maximum

            parsed_name_match = re.match(AMOUNT_CAPABILITY_NAME_REGEX, self.name)

            if parsed_name_match and parsed_name_match.group(1) in RESERVED_FIRST_IDENTIFIERS:
                raise NonValidInputError(
                    "Please make sure that the first identifier in your name is not a reserved identifier. "
                    + str(RESERVED_FIRST_IDENTIFIERS)
                )
        else:
            raise NonValidInputError(
                "Please fill out all custom amount names in the custom host requirement options!"
            )
        return requirement


class CustomAttributeWidget(CustomCapabilityWidget):
    """
    UI element to hold a single custom attribute.
    """

    ANY_OF = "anyOf"
    ALL_OF = "allOf"

    def __init__(
        self, list_item: QListWidgetItem, item_number: int, parent=CustomRequirementsWidget
    ):
        super().__init__(ATTRIBUTE, list_item, item_number, parent)
        self._build_ui()

    def _build_ui(self):
        # Name / Value / All / Any
        self.name_label = QLabel("Attribute name")
        self.name_label.setFixedWidth(LABEL_FIXED_WIDTH)
        self.value_label = QLabel("Value(s)")
        self.all_of_button = QRadioButton("All")
        self.all_of_button.setChecked(True)
        self.any_of_button = QRadioButton("Any")
        self.name_line_edit = QLineEdit()
        self.name_line_edit.setFixedWidth(LABEL_FIXED_WIDTH)
        assert (100 - len(ATTRIBUTE_CAPABILITY_PREFIX)) > 0
        self.name_line_edit.setMaxLength(100 - len(ATTRIBUTE_CAPABILITY_PREFIX))
        self.name_line_edit.setValidator(
            QRegularExpressionValidator(ATTRIBUTE_CAPABILITY_NAME_REGEX)
        )
        self.add_value_button = None

        self.top_row = QHBoxLayout()
        self.top_row.addWidget(self.value_label)

        self.top_row.addStretch()
        self.top_row.addWidget(self.all_of_button)
        self.top_row.addWidget(self.any_of_button)

        # Create a list widget for placing custom attribute values
        self.value_list_widget = QListWidget(self)
        self.value_list_widget.setSelectionMode(QListView.NoSelection)
        self.value_list_widget.viewport().setAutoFillBackground(False)
        self.value_list_widget.setFrameStyle(QFrame.NoFrame)
        self.value_list_widget.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.value_list_widget.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.value_list_widget.setSizeAdjustPolicy(QListWidget.AdjustToContents)

        self.name_column = QVBoxLayout()
        self.name_column.setContentsMargins(2, 0, 0, 0)
        self.name_column.setAlignment(Qt.AlignTop)
        self.name_column.addWidget(self.name_label)
        self.name_column.addWidget(self.name_line_edit)

        self.value_column = QVBoxLayout()
        self.value_column.setContentsMargins(0, 0, 0, 0)
        self.value_column.addLayout(self.top_row)
        self.value_column.addWidget(self.value_list_widget)

        # Reduce the spacing between top row and value list
        self.value_column.setSpacing(2)

        self.columns_widget = QWidget(self)
        self.value_column_widget = QWidget(self.columns_widget)
        self.value_column_widget.setLayout(self.value_column)
        self.columns = QVBoxLayout()
        self.columns.setContentsMargins(0, 15, 0, 0)
        self.columns.addLayout(self.name_column)
        self.columns.addWidget(self.value_column_widget)

        # LineEdit  / LineEdit / Optional [X]
        self.columns_widget.setLayout(self.columns)
        self.layout.addWidget(self.columns_widget)
        self._add_value()

    def _add_value(self):  # type: ignore
        # Return type CustomAttributeValueWidget is not yet defined
        value_list_item = QListWidgetItem(self.value_list_widget)
        value = CustomAttributeValueWidget(value_list_item, self)
        value_list_item.setSizeHint(value.sizeHint())
        self.value_list_widget.addItem(value_list_item)
        self.value_list_widget.setItemWidget(value_list_item, value)
        self._resize_value_list_to_fit(1)
        self._move_add_button_to_last_item()
        self._set_remove_button_for_first_item()
        return value

    def remove_value_item(self, value):
        # remove the ListWidgetItem from value_list_item
        self.value_list_widget.takeItem(self.value_list_widget.indexFromItem(value).row())
        self._resize_value_list_to_fit(-1)
        self._move_add_button_to_last_item()
        self._set_remove_button_for_first_item()

    def _resize_value_list_to_fit(self, item_count_change: int):
        # Resize the list widget as well as parents to based on the size of the contents
        self.value_column_widget.setFixedHeight(
            self.value_column_widget.height()
            + item_count_change
            * self.value_list_widget.itemWidget(self.value_list_widget.item(0)).height()
        )
        self.columns_widget.adjustSize()
        self.adjustSize()
        self.list_item.setSizeHint(self.sizeHint())
        self._parent.resize_list_to_fit()

    def _move_add_button_to_last_item(self):
        # Add value button
        if self.add_value_button is not None:
            self.add_value_button.setParent(None)

        else:
            self.add_value_button = QPushButton("Add")
            self.add_value_button.setStyleSheet("border-width: 0px")
            self.add_value_button.setToolTip(
                "Add a new value to evaluate against for this attribute"
            )
            self.add_value_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

            self.add_value_button.clicked.connect(self._add_value)

        last_item = self.value_list_widget.itemWidget(
            self.value_list_widget.item(self.value_list_widget.count() - 1)
        )
        last_item.layout.insertWidget(last_item.layout.count() - 2, self.add_value_button)

    def _set_remove_button_for_first_item(self):
        if self.value_list_widget.count() == 1:
            first_item = self.value_list_widget.itemWidget(self.value_list_widget.item(0))
            first_item.remove_button.setEnabled(False)
            first_item.remove_button.unsetCursor()

        if self.value_list_widget.count() >= 2:
            first_item = self.value_list_widget.itemWidget(self.value_list_widget.item(0))
            first_item.remove_button.setEnabled(True)

    @property
    def name(self) -> str:
        return self.name_line_edit.text()

    @name.setter
    def name(self, name: str):
        self.name_line_edit.setText(name)

    @property
    def option(self) -> Literal[ANY_OF, ALL_OF]:  # type: ignore
        return self.ANY_OF if self.any_of_button.isChecked() else self.ALL_OF

    @option.setter
    def option(self, option: Literal[ANY_OF, ALL_OF]):  # type: ignore
        if option == self.ANY_OF:
            self.all_of_button.setChecked(False)
            self.any_of_button.setChecked(True)
        else:
            self.all_of_button.setChecked(True)
            self.any_of_button.setChecked(False)

    @property
    def values(self) -> List[str]:
        values = []
        for i in range(self.value_list_widget.count()):
            value: CustomAttributeValueWidget = self.value_list_widget.itemWidget(
                self.value_list_widget.item(i)
            )
            if value.line_edit.text():
                values.append(value.line_edit.text())
            else:
                raise ValueError("All values must have input")
        return values

    @values.setter
    def values(self, values: List[str]):
        value_widgets_by_value = {}
        for i in range(self.value_list_widget.count()):
            value_widget: CustomAttributeValueWidget = self.value_list_widget.itemWidget(
                self.value_list_widget.item(i)
            )
            if value_widget.line_edit.text():
                value_widgets_by_value[value_widget.line_edit.text()] = value_widget
            else:
                value_widgets_by_value["EMPTY"] = value_widget

        has_only_empty_value = False
        if (
            len(value_widgets_by_value.keys()) == 1
            and list(value_widgets_by_value.keys())[0] == "EMPTY"
        ):
            has_only_empty_value = True

        for value in values:
            if has_only_empty_value:
                value_widget = value_widgets_by_value["EMPTY"]
                value_widget.line_edit.setText(value)
                value_widgets_by_value[value] = value_widget
                del value_widgets_by_value["EMPTY"]
                has_only_empty_value = False
            if value in value_widgets_by_value:
                value_widget = value_widgets_by_value[value]
                value_widget.line_edit.setText(value)
            else:
                value_widget = self._add_value()
                value_widget.line_edit.setText(value)

    def get_requirement(self) -> Dict[str, Any]:
        """
        Return an OpenJD parameter definition dict with
        a "value" key filled from the widget, and a list of values.

        An attribute capability is prefixed with "attr.worker".
        """
        requirement: Dict[str, Any] = {}
        requirements_are_valid = True

        if self.name_line_edit.text():
            if self.name[-1] == ".":
                raise NonValidInputError("Your requirement name cannot end with a period.")

            parsed_name_match = re.match(AMOUNT_CAPABILITY_NAME_REGEX, self.name)
            if parsed_name_match and parsed_name_match.group(1) in RESERVED_FIRST_IDENTIFIERS:
                raise NonValidInputError(
                    "Please make sure that the first identifier in your name is not a reserved identifier. "
                    + str(RESERVED_FIRST_IDENTIFIERS)
                )

            try:
                values = self.values
            except ValueError:
                values = []
                requirements_are_valid = False
            if values:
                requirement = {
                    "name": ATTRIBUTE_CAPABILITY_PREFIX + self.name_line_edit.text(),
                    f"{self.option}": values,
                }
        else:
            requirements_are_valid = False

        if not requirements_are_valid:
            raise NonValidInputError(
                "Please fill out all custom attribute names and values in the custom host requirements options!"
            )

        return requirement


class CustomAttributeValueWidget(QWidget):
    """
    UI element to hold a single custom attribute value.
    """

    def __init__(self, value_list_item: QListWidgetItem, parent: CustomAttributeWidget):
        super().__init__(parent)
        self._parent: CustomAttributeWidget = parent
        self.value_list_item = value_list_item

        self.line_edit = QLineEdit()
        self.line_edit.setFixedWidth(LABEL_FIXED_WIDTH + 20)
        self.line_edit.setMaxLength(100)
        self.line_edit.setValidator(QRegularExpressionValidator(ATTRIBUTE_CAPABILITY_VALUE_REGEX))

        self.remove_button = QPushButton("Remove")
        self.remove_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.remove_button.clicked.connect(self._remove)

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(2, 0, 0, 0)
        self.layout.addWidget(self.line_edit)
        self.layout.addWidget(self.remove_button)
        self.layout.addStretch()
        self.layout.setAlignment(Qt.AlignLeft)

    def _remove(self):
        self._parent.remove_value_item(self.value_list_item)
        self.setParent(None)
        self.deleteLater()


class OSRequirementRowWidget(QWidget):
    """
    UI element to hold a single row of OS requirement components.

    Args:
        label: the name of the requirement
        items: selectable options to display inside a comboBox
        parent: The parent Qt Widget
    """

    def __init__(self, label: str, items: List[str], parent=None):
        super().__init__(parent)
        self.layout = QHBoxLayout(self)
        # remove default spaces around BoxLayout
        self.layout.setContentsMargins(0, 0, 0, 0)
        self._build_ui(label, items)

    def _build_ui(self, label: str, items: List[str]):
        self.label = QLabel(label)
        self.label.setFixedWidth(LABEL_FIXED_WIDTH)
        self.combo_box = OptionalMultiSelectComboBox(items, parent=self)
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.combo_box)


class HardwareRequirementsRowWidget(QWidget):
    """
    UI element to hold a single row of hardware requirement components.

    Args:
        label: the name of the requirement
        parent: The parent Qt Widget
    """

    def __init__(self, label: str, parent=None):
        super().__init__(parent)
        self.layout = QHBoxLayout(self)
        # remove default spaces around BoxLayout
        self.layout.setContentsMargins(0, 0, 0, 0)
        self._build_ui(label)

    def _build_ui(self, label: str):
        self.label = QLabel(label)
        self.label.setFixedWidth(LABEL_FIXED_WIDTH)

        # Create "Min" label, and set label to fixed width
        self.min_label = QLabel("Min")
        self.min_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        # Create "Max" label, and set label to fixed width
        self.max_label = QLabel("Max")
        self.max_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        # Create spin box for "Min"
        self.min_spin_box = OptionalSpinBox(min=0, max=100000, parent=self)

        # Create spin box for "Max"
        self.max_spin_box = OptionalSpinBox(min=0, max=100000, parent=self)

        self.layout.addWidget(self.label)
        self.layout.addWidget(self.min_label)
        self.layout.addWidget(self.min_spin_box)
        self.layout.addWidget(self.max_label)
        self.layout.addWidget(self.max_spin_box)

    def add_requirement(self, requirements: List, name: str, scaling_factor: int = 1):
        """
        Create a dict based on whether inputs have been received for the components in the row widget,
        then append the dict to an existing list of requirements.

        Args:
            requirements: the list of requirements to append to
            name: the name of the capability
            scaling_factor: for some of the amount capabilities, the unit displayed on the UI is different
                then the unit used within template, so use this factor to scale the input values.
        """
        if self.min_spin_box.has_input() or self.max_spin_box.has_input():
            requirement = {"name": name}
            if self.min_spin_box.has_input():
                requirement["min"] = self.min_spin_box.value() * scaling_factor
            if self.max_spin_box.has_input():
                requirement["max"] = self.max_spin_box.value() * scaling_factor
            requirements.append(requirement)

    def clear(self):
        self.min_spin_box.setValue(self.min_spin_box.no_input_value)
        self.max_spin_box.setValue(self.max_spin_box.no_input_value)

    @property
    def minimum(self):
        return self.min_spin_box.value()

    @minimum.setter
    def minimum(self, minimum_value: int):
        if minimum_value == self.min_spin_box.no_input_value:
            self.min_spin_box.setValue(self.min_spin_box.no_input_value)
        else:
            self.min_spin_box.setValue(minimum_value)

    @property
    def maximum(self):
        return self.max_spin_box.value()

    @maximum.setter
    def maximum(self, maximum_value: int):
        if maximum_value == self.max_spin_box.no_input_value:
            self.max_spin_box.setValue(self.max_spin_box.no_input_value)
        else:
            self.max_spin_box.setValue(maximum_value)


class OptionalComboBox(QComboBox):
    """
    A custom QComboBox that always have a grayed out placeholder option.
    """

    def __init__(self, items: List[str], parent=None):
        super().__init__(parent)
        self.addItem(PLACEHOLDER_TEXT)
        self.setItemData(0, QBrush(Qt.gray), Qt.TextColorRole)
        self.addItems(items)

    def has_input(self) -> bool:
        return self.currentText() != PLACEHOLDER_TEXT


class OptionalMultiSelectComboBox(QComboBox):
    """
    A custom QComboBox that always have a grayed out placeholder option and supports multiselect.
    """

    def __init__(self, items: List[str], parent=None):
        super().__init__(parent=parent)
        self.model().itemChanged.connect(self.handleModelChanged)
        self.setPlaceholderText("-")
        self.addItems(items)
        self._updateLineEdit()

    def handleModelChanged(self, item):
        self._updateLineEdit()

    def _updateLineEdit(self):
        text = ", ".join(self.checkedItems())
        self.setPlaceholderText(text or "-")

    def setChecked(self, items: List[str]):  # type: ignore
        for item in items:
            for i in range(self.model().rowCount()):
                model_item = self.model().item(i)
                if model_item.text().lstrip(" ") == item:
                    model_item.setData(Qt.Checked, Qt.CheckStateRole)

    def checkedItems(self):
        items = []
        for i in range(self.model().rowCount()):
            item = self.model().item(i)
            if item.checkState() == Qt.Checked:
                items.append(item.text().lstrip(" "))
        return items

    def addItems(self, items):
        for item in items:
            model_item = QStandardItem("    " + item)
            model_item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
            model_item.setData(Qt.Unchecked, Qt.CheckStateRole)
            self.model().appendRow([model_item])


class OptionalSpinBox(QSpinBox):
    """
    A custom QSpinBox that set min - 1 value as "-" to represent value not set.
    """

    def __init__(self, min: int = MIN_INT_VALUE, max: int = MAX_INT_VALUE, parent=None) -> None:
        super().__init__(parent)
        self.min = min
        self.max = max
        self.no_input_value = min - 1
        # Set the range to include min-1 as a valid value
        self.setRange(self.no_input_value, MAX_INT_VALUE)
        self.setValue(self.no_input_value)

    def validate(self, input: str, pos: int) -> QValidator.State:
        """
        Override validate function to treat empty string "" as acceptable.
        """
        if input == "" or input == PLACEHOLDER_TEXT:
            return QValidator.Acceptable
        else:
            # Override the range that can be accepted by the validator by actual min/max values.
            validator = QIntValidator()
            validator.setBottom(self.min)
            validator.setTop(self.max)
            return validator.validate(input, pos)  # type: ignore[return-value]

    def valueFromText(self, text: str) -> int:
        """
        Override valueFromText function to return no-input-value if input string is empty or placeholder.
        """
        if text == "" or text == PLACEHOLDER_TEXT:
            return self.no_input_value
        else:
            return super().valueFromText(text)

    def textFromValue(self, val: int) -> str:
        """
        Override textFromValue function to return placeholder text if value is no-input-value.
        """
        if val == self.no_input_value:
            return PLACEHOLDER_TEXT
        else:
            return super().textFromValue(val)

    def wheelEvent(self, event):
        """
        Override wheelEvent to disable scrolling from accidentally gaining focus and changing the numbers.
        """
        event.ignore()

    def has_input(self) -> bool:
        """
        Custom function to indicate whether the SpinBox has received input.
        """
        return self.no_input_value != self.value()

    def stepBy(self, steps: int) -> None:
        current_value: int = self.value()
        result_value = self.value() + steps
        if (
            result_value == self.no_input_value
            or result_value > self.maximum()
            or result_value < self.minimum()
        ):
            # If result value is not a valid value, do not go to that value
            return

        if (
            current_value == self.no_input_value and steps == 1
        ):  # We should allow the user to increment from null value to a valid value
            super().setValue(max(self.min, 0))
        else:
            super().stepBy(steps)


class OptionalDoubleSpinBox(QDoubleSpinBox):
    """
    A custom QDoubleSpinBox that set min - 1 value as "-" to represent value not set.
    """

    def __init__(
        self,
        min: int = MIN_INT_VALUE,
        max: int = MAX_INT_VALUE,
        decimal: int = DECIMAL_VALUE,
        parent=None,
    ) -> None:
        super().__init__(parent)
        self.min = min
        self.max = max
        self.no_input_value = min - 1
        self.decimal = decimal
        # Set the range to include min-1 as a valid value
        self.setRange(self.no_input_value, MAX_INT_VALUE)
        self.setValue(self.no_input_value)
        self.setDecimals(decimal)

    def validate(self, input: str, pos: int) -> QValidator.State:
        """
        Override validate function to treat empty string "" as acceptable.
        """
        if input == "" or input == PLACEHOLDER_TEXT:
            return QValidator.Acceptable
        else:
            # Override the range that can be accepted by the validator by actual min/max values.
            validator = QDoubleValidator()
            validator.setBottom(self.min)
            validator.setTop(self.max)
            validator.setDecimals(self.decimal)
            return validator.validate(input, pos)  # type: ignore[return-value]

    def valueFromText(self, text: str) -> float:
        """
        Override valueFromText function to return no-input-value if input string is empty or placeholder.
        """
        if text == "" or text == PLACEHOLDER_TEXT:
            return self.no_input_value
        else:
            return super().valueFromText(text)

    def textFromValue(self, val: float) -> str:
        """
        Override textFromValue function to return placeholder text if value is no-input-value.
        """
        if val == self.no_input_value:
            return PLACEHOLDER_TEXT
        else:
            return super().textFromValue(val)

    def wheelEvent(self, event):
        """
        Override wheelEvent to disable scrolling from accidentally gaining focus and changing the numbers.
        """
        event.ignore()

    def has_input(self) -> bool:
        """
        Custom function to indicate whether the SpinBox has received input.
        """
        return self.no_input_value != self.value()

    def stepBy(self, steps: int) -> None:
        current_value: int = self.value()
        result_value = self.value() + steps
        if (
            result_value == self.no_input_value
            or result_value > self.maximum()
            or result_value < self.minimum()
        ):
            # If result value is not a valid value, do not go to that value
            return

        if (
            current_value == self.no_input_value and steps == 1
        ):  # We should allow the user to increment from null value to a valid value
            super().setValue(max(self.min, 0))
        else:
            super().stepBy(steps)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import datetime
from io import BytesIO
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import boto3

from deadline.client.api._session import _get_queue_user_boto3_session, get_default_client_config
from deadline.client.cli._groups.click_logger import ClickLogger
from deadline.job_attachments._diff import _fast_file_list_to_manifest_diff, compare_manifest
from deadline.job_attachments._glob import _process_glob_inputs, _glob_paths
from deadline.job_attachments.api._utils import _read_manifests
from deadline.job_attachments.asset_manifests._create_manifest import (
    _create_manifest_for_single_root,
)
from deadline.job_attachments.asset_manifests.base_manifest import (
    BaseAssetManifest,
    BaseManifestPath,
)
from deadline.client.config import config_file
from deadline.job_attachments.asset_manifests.decode import decode_manifest
from deadline.job_attachments.asset_manifests.hash_algorithms import hash_data
from deadline.job_attachments.caches.hash_cache import HashCache
from deadline.job_attachments.download import (
    get_manifest_from_s3,
    get_output_manifests_by_asset_root,
    merge_asset_manifests,
)
from deadline.job_attachments.models import (
    S3_MANIFEST_FOLDER_NAME,
    FileStatus,
    GlobConfig,
    JobAttachmentS3Settings,
    ManifestDiff,
    ManifestDownload,
    ManifestDownloadResponse,
    ManifestSnapshot,
    ManifestMerge,
    default_glob_all,
    AssetType,
)
from deadline.job_attachments._utils import _get_long_path_compatible_path
from deadline.job_attachments.upload import S3AssetManager, S3AssetUploader

"""
APIs here should be business logic only. It should perform one thing, and one thing well. 
It should use basic primitives like S3 upload, download, boto3 APIs.
These APIs should be boto3 session agnostic and a specific Boto3 Credential to use.
"""


def _glob_files(
    root: str,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    include_exclude_config: Optional[str] = None,
) -> List[str]:
    """
    :param include: Include glob to look for files to add to the manifest.
    :param exclude: Exclude glob to exclude files from the manifest.
    :param include_exclude_config: Config JSON or file containeing input and exclude config.
    :returns: All files matching the include and exclude expressions.
    """

    # Get all files in the root.
    glob_config: GlobConfig
    if include or exclude:
        include = include if include is not None else default_glob_all()
        exclude = exclude if exclude is not None else []
        glob_config = GlobConfig(include_glob=include, exclude_glob=exclude)
    elif include_exclude_config:
        glob_config = _process_glob_inputs(include_exclude_config)
    else:
        # Default, include all.
        glob_config = GlobConfig()

    input_files = _glob_paths(
        root, include=glob_config.include_glob, exclude=glob_config.exclude_glob
    )
    return input_files


def _manifest_snapshot(
    root: str,
    destination: str,
    name: str,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    include_exclude_config: Optional[str] = None,
    diff: Optional[str] = None,
    force_rehash: bool = False,
    logger: ClickLogger = ClickLogger(False),
) -> Optional[ManifestSnapshot]:
    # Get all files in the root.
    glob_config: GlobConfig
    if include or exclude:
        include = include if include is not None else default_glob_all()
        exclude = exclude if exclude is not None else []
        glob_config = GlobConfig(include_glob=include, exclude_glob=exclude)
    elif include_exclude_config:
        glob_config = _process_glob_inputs(include_exclude_config)
    else:
        # Default, include all.
        glob_config = GlobConfig()

    current_files = _glob_paths(
        root, include=glob_config.include_glob, exclude=glob_config.exclude_glob
    )

    # Compute the output manifest immediately and hash.
    if not diff:
        output_manifest = _create_manifest_for_single_root(
            files=current_files, root=root, logger=logger
        )
        if not output_manifest:
            return None

    # If this is a diff manifest, load the supplied manifest file.
    else:
        # Parse local manifest
        with open(diff) as source_diff:
            source_manifest_str = source_diff.read()
            source_manifest = decode_manifest(source_manifest_str)

        # Get the differences
        changed_paths: List[str] = []

        # Fast comparison using time stamps and sizes.
        if not force_rehash:
            diff_list: List[Tuple[str, FileStatus]] = _fast_file_list_to_manifest_diff(
                root=root,
                current_files=current_files,
                diff_manifest=source_manifest,
                logger=logger,
                return_root_relative_path=False,
            )
            for diff_file in diff_list:
                # Add all new and modified
                if diff_file[1] != FileStatus.DELETED:
                    changed_paths.append(diff_file[0])
        else:
            # In "slow / thorough" mode, we check by hash, which is definitive.
            output_manifest = _create_manifest_for_single_root(
                files=current_files, root=root, logger=logger
            )
            if not output_manifest:
                return None
            differences: List[Tuple[FileStatus, BaseManifestPath]] = compare_manifest(
                source_manifest, output_manifest
            )
            for diff_item in differences:
                if diff_item[0] == FileStatus.MODIFIED or diff_item[0] == FileStatus.NEW:
                    full_diff_path = f"{root}/{diff_item[1].path}"
                    changed_paths.append(full_diff_path)
                    logger.echo(f"Found difference at: {full_diff_path}, Status: {diff_item[0]}")

        # If there were no files diffed, return None, there was nothing to snapshot.
        if len(changed_paths) == 0:
            return None

        # Since the files are already hashed, we can easily re-use has_attachments to remake a diff manifest.
        output_manifest = _create_manifest_for_single_root(
            files=changed_paths, root=root, logger=logger
        )
        if not output_manifest:
            return None

    # Write created manifest into local file, at the specified location at destination
    if output_manifest is not None:
        local_manifest_file = _write_manifest(
            root=root,
            manifest=output_manifest,
            destination=destination,
            name=name,
        )
        # Output results.
        logger.echo(f"Manifest generated at {local_manifest_file}")
        return ManifestSnapshot(root=root, manifest=local_manifest_file)
    else:
        # No manifest generated.
        logger.echo("No manifest generated")
        return None


def _write_manifest(
    root: str,
    manifest: BaseAssetManifest,
    destination: str,
    name: Optional[str] = None,
) -> str:
    """
    Write a manifest to a destination.
    """
    # Write created manifest into local file, at the specified location at destination
    root_hash: str = hash_data(root.encode("utf-8"), manifest.get_default_hash_alg())
    timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
    manifest_name = name if name else root.replace("/", "_").replace("\\", "_").replace(":", "_")
    manifest_name = manifest_name[1:] if manifest_name[0] == "_" else manifest_name
    manifest_name = f"{manifest_name}-{root_hash}-{timestamp}.manifest"

    local_manifest_path = str(
        _get_long_path_compatible_path(
            os.path.join(destination, manifest_name),
        )
    )
    os.makedirs(os.path.dirname(local_manifest_path), exist_ok=True)
    with open(local_manifest_path, "w") as file:
        file.write(manifest.encode())

    return local_manifest_path


def _manifest_diff(
    manifest: str,
    root: str,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    include_exclude_config: Optional[str] = None,
    force_rehash=False,
    logger: ClickLogger = ClickLogger(False),
) -> ManifestDiff:
    """
    BETA API - This API is still evolving but will be made public in the near future.
    API to diff a manifest root with a previously snapshotted manifest.
    :param manifest: Manifest file path to compare against.
    :param root: Root directory to generate the manifest fileset.
    :param include: Include glob to look for files to add to the manifest.
    :param exclude: Exclude glob to exclude files from the manifest.
    :param include_exclude_config: Config JSON or file containeing input and exclude config.
    :param logger: Click Logger instance to print to CLI as text or JSON.
    :returns: ManifestDiff object containing all new changed, deleted files.
    """

    # Find all files matching our regex
    input_files = _glob_files(
        root=root, include=include, exclude=exclude, include_exclude_config=include_exclude_config
    )
    input_paths = [Path(p) for p in input_files]

    # Placeholder Asset Manager
    asset_manager = S3AssetManager()

    # parse the given manifest to compare against.
    local_manifest_object: BaseAssetManifest
    with open(manifest) as input_file:
        manifest_data_str = input_file.read()
        local_manifest_object = decode_manifest(manifest_data_str)

    output: ManifestDiff = ManifestDiff()

    # Helper function to update output datastructure.
    def process_output(status: FileStatus, path: str, output_diff: ManifestDiff):
        if status == FileStatus.MODIFIED:
            output_diff.modified.append(path)
        elif status == FileStatus.NEW:
            output_diff.new.append(path)
        elif status == FileStatus.DELETED:
            output_diff.deleted.append(path)

    if force_rehash:
        # hash and create manifest of local directory
        cache_config = config_file.get_cache_directory()
        with HashCache(cache_config) as hash_cache:
            directory_manifest_object = asset_manager._create_manifest_file(
                input_paths=input_paths, root_path=root, hash_cache=hash_cache
            )

        # Hash based compare manifests.
        differences: List[Tuple[FileStatus, BaseManifestPath]] = compare_manifest(
            reference_manifest=local_manifest_object, compare_manifest=directory_manifest_object
        )
        # Map to output datastructure.
        for item in differences:
            process_output(item[0], item[1].path, output)

    else:
        # File based comparisons.
        fast_diff: List[Tuple[str, FileStatus]] = _fast_file_list_to_manifest_diff(
            root=root, current_files=input_files, diff_manifest=local_manifest_object, logger=logger
        )
        for fast_diff_item in fast_diff:
            process_output(fast_diff_item[1], fast_diff_item[0], output)

    return output


def _manifest_upload(
    manifest_file: str,
    s3_bucket_name: str,
    s3_cas_prefix: str,
    boto_session: boto3.Session,
    s3_key_prefix: Optional[str] = None,
    logger: ClickLogger = ClickLogger(False),
):
    """
    BETA API - This API is still evolving but will be made public in the near future.
    API to upload a job attachment manifest to the Content Addressable Storage. Manifests will be
    uploaded to s3://{s3_bucket_name}/{cas_prefix}/Manifests/{s3_key_prefix}/{manifest_file_name} as per the Deadline CAS folder structure.
    manifest_file: File Path to the manifest file for upload.
    s3_bucket_name: S3 bucket name.
    boto_session: S3 Content Addressable Storage prefix.
    s3_key_prefix: [Optional] S3 prefix path to the Content Addressable Storge.
    boto_session: Boto3 session.
    logger: Click Logger instance to print to CLI as text or JSON.
    """
    # S3 metadata

    # Upload settings:
    s3_metadata: Dict[str, Any] = {"Metadata": {}}
    s3_metadata["Metadata"]["file-system-location-name"] = manifest_file

    # Always upload the manifest file to case root /Manifest with the original file name.
    manifest_path: str = "/".join(
        [s3_cas_prefix, S3_MANIFEST_FOLDER_NAME, s3_key_prefix, Path(manifest_file).name]
        if s3_key_prefix
        else [s3_cas_prefix, S3_MANIFEST_FOLDER_NAME, Path(manifest_file).name]
    )

    # S3 uploader.
    upload = S3AssetUploader(session=boto_session)

    manifest_file = str(_get_long_path_compatible_path(manifest_file))

    with open(manifest_file) as manifest:
        upload.upload_bytes_to_s3(
            bytes=BytesIO(manifest.read().encode("utf-8")),
            bucket=s3_bucket_name,
            key=manifest_path,
            progress_handler=logger.echo,
            extra_args=s3_metadata,
        )


def _manifest_download(
    download_dir: str,
    farm_id: str,
    queue_id: str,
    job_id: str,
    boto3_session: boto3.Session,
    step_id: Optional[str] = None,
    asset_type: AssetType = AssetType.ALL,
    logger: ClickLogger = ClickLogger(False),
) -> ManifestDownloadResponse:
    """
    BETA API - This API is still evolving but will be made public in the near future.
    API to download the Job Attachment manifest for a Job, and optionally dependencies for Step.
    download_dir: Download directory.
    farm_id: The Deadline Farm to download from.
    queue_id: The Deadline Queue to download from.
    job_id: Job Id to download.
    boto_session: Boto3 session.
    step_id: Optional[str]: Optional, download manifest for a step
    asset_type: Which asset manifests should be downloaded for given job (& optionally step), options are Input, Output, All. Default behaviour is All.
    logger: Click Logger instance to print to CLI as text or JSON.
    return ManifestDownloadResponse Downloaded Manifest data. Contains source S3 key and local download path.
    """

    # Deadline Client and get the Queue to download.
    deadline = boto3_session.client("deadline", config=get_default_client_config())

    queue: dict = deadline.get_queue(
        farmId=farm_id,
        queueId=queue_id,
    )

    # assume queue role - session permissions
    queue_role_session: boto3.Session = _get_queue_user_boto3_session(
        deadline=deadline,
        base_session=boto3_session,
        farm_id=farm_id,
        queue_id=queue_id,
        queue_display_name=queue["displayName"],
    )

    # Queue's Job Attachment settings.
    queue_s3_settings = JobAttachmentS3Settings(**queue["jobAttachmentSettings"])

    # Get S3 prefix
    s3_prefix: Path = Path(queue_s3_settings.rootPrefix, S3_MANIFEST_FOLDER_NAME)

    # Capture a list of success download files for JSON output.
    successful_downloads: List[ManifestDownload] = []

    # Utility function to build up manifests by root.
    manifests_by_root: Dict[str, List[BaseAssetManifest]] = dict()

    # Set the values of download input & output as per selected asset types in the api request
    download_input: bool = (
        True if asset_type is None or asset_type in (AssetType.INPUT, AssetType.ALL) else False
    )
    download_output: bool = (
        True if asset_type is None or asset_type in (AssetType.OUTPUT, AssetType.ALL) else False
    )

    def add_manifest_by_root(
        manifests_by_root: Dict[str, list], root: str, manifest: BaseAssetManifest
    ):
        if root not in manifests_by_root:
            manifests_by_root[root] = []
        manifests_by_root[root].append(manifest)

    # Get the job from deadline api
    job: dict = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)

    # If input manifests need to be downloaded
    if download_input:
        logger.echo(f"Downloading input manifests for job: {job_id}")

        # Get input_manifest_paths from Deadline GetJob API
        attachments: dict = job.get("attachments", {})
        input_manifest_paths: List[Tuple[str, str]] = [
            (manifest.get("inputManifestPath", ""), manifest["rootPath"])
            for manifest in attachments.get("manifests", [])
        ]

        # Download each input_manifest_path
        for input_manifest_path, root_path in input_manifest_paths:
            asset_manifest: BaseAssetManifest = get_manifest_from_s3(
                manifest_key=(s3_prefix / input_manifest_path).as_posix(),
                s3_bucket=queue_s3_settings.s3BucketName,
                session=queue_role_session,
            )
            if asset_manifest is not None:
                logger.echo(f"Found input manifest for root: {root_path}")
                add_manifest_by_root(
                    manifests_by_root=manifests_by_root, root=root_path, manifest=asset_manifest
                )

        # Now handle step-step dependencies
        if step_id is not None:
            logger.echo(f"Finding step-step dependency manifests for step: {step_id}")

            # Get Step-Step dependencies with pagination
            next_token = ""
            while next_token is not None:
                step_dep_response = deadline.list_step_dependencies(
                    farmId=farm_id,
                    queueId=queue_id,
                    jobId=job_id,
                    stepId=step_id,
                    nextToken=next_token,
                )

                for dependent_step in step_dep_response["dependencies"]:
                    logger.echo(f"Found Step-Step dependency. {dependent_step['stepId']}")

                    # Get manifests for the step-step dependency
                    step_manifests_by_root: Dict[str, List[BaseAssetManifest]] = (
                        get_output_manifests_by_asset_root(
                            s3_settings=queue_s3_settings,
                            farm_id=farm_id,
                            queue_id=queue_id,
                            job_id=job_id,
                            step_id=dependent_step["stepId"],
                            session=queue_role_session,
                        )
                    )
                    # Merge all manifests by root.
                    for root in step_manifests_by_root.keys():
                        for manifest in step_manifests_by_root[root]:
                            logger.echo(f"Found step-step output manifest for root: {root}")
                            add_manifest_by_root(
                                manifests_by_root=manifests_by_root, root=root, manifest=manifest
                            )

                next_token = step_dep_response.get("nextToken")

    # If output manifests need to be downloaded
    if download_output:
        output_manifests_by_root: Dict[str, List[BaseAssetManifest]]
        if step_id is not None:
            logger.echo(f"Downloading output manifests step: {step_id} of job: {job_id}")
            # Only get the output manifests for selected step
            output_manifests_by_root = get_output_manifests_by_asset_root(
                s3_settings=queue_s3_settings,
                farm_id=farm_id,
                queue_id=queue_id,
                job_id=job_id,
                step_id=step_id,
                session=queue_role_session,
            )

        else:
            logger.echo(f"Downloading output manifests for job: {job_id}")
            # Get output manifests for all steps of the job
            output_manifests_by_root = get_output_manifests_by_asset_root(
                s3_settings=queue_s3_settings,
                farm_id=farm_id,
                queue_id=queue_id,
                job_id=job_id,
                session=queue_role_session,
            )

        # Merge all output manifests by root.
        for root in output_manifests_by_root.keys():
            for manifest in output_manifests_by_root[root]:
                logger.echo(f"Found output manifest for root: {root}")
                add_manifest_by_root(
                    manifests_by_root=manifests_by_root, root=root, manifest=manifest
                )

    # Finally, merge all manifest paths to create unified manifests.
    # TODO: Filter outputs by path

    merged_manifests: Dict[str, BaseAssetManifest] = {}
    for root in manifests_by_root.keys():
        merged_manifest = merge_asset_manifests(manifests_by_root[root])
        if merged_manifest:
            merged_manifests[root] = merged_manifest

    # Save the manifest files to disk.
    for root in merged_manifests.keys():
        # Save the merged manifest as {root}_{hash}_timestamp.
        root_hash: str = hash_data(
            root.encode("utf-8"), merged_manifests[root].get_default_hash_alg()
        )
        timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
        manifest_name = root.replace("/", "_")
        manifest_name = manifest_name[1:] if manifest_name[0] == "_" else manifest_name
        manifest_name = f"{manifest_name}-{root_hash}-{timestamp}.manifest"

        local_manifest_file_path = os.path.join(download_dir, manifest_name)
        with open(local_manifest_file_path, "w") as file:
            file.write(merged_manifests[root].encode())
        successful_downloads.append(
            ManifestDownload(manifest_root=root, local_manifest_path=str(local_manifest_file_path))
        )
        logger.echo(f"Downloaded merged manifest for root: {root} to: {local_manifest_file_path}")

    # JSON output at the end.
    output = ManifestDownloadResponse(downloaded=successful_downloads)
    return output


def _manifest_merge(
    root: str,
    manifest_files: List[str],
    destination: str,
    name: Optional[str],
    logger: ClickLogger = ClickLogger(False),
) -> Optional[ManifestMerge]:
    """
    BETA API - API to merge multiple manifests into one.
    root: Root path for the manifest.
    manifest_files: List of manifest files to merge.
    destination: Destination directory for the merged manifest.
    name: Name of the merged manifest.
    logger: Click Logger instance to print to CLI as text or JSON.
    return ManifestMerge object containing the merged manifest.
    """

    manifests: List[BaseAssetManifest] = list(
        _read_manifests(manifest_paths=manifest_files).values()
    )

    merged_manifest = merge_asset_manifests(manifests)

    if not merged_manifest:
        return None

    local_manifest_file = _write_manifest(
        root=root, manifest=merged_manifest, destination=destination, name=name
    )
    logger.echo(f"Manifest generated at {local_manifest_file}")

    return ManifestMerge(manifest_root=root, local_manifest_path=local_manifest_file)
